"""
Insert Raster
=============

The `RasterElement` objects store the Raster data in WKB form. This WKB format is usually fetched
from the database but when the data comes from another source it can be hard to format it as as a
WKB. This example shows a method to convert input data into a WKB in order to insert it.
This example uses SQLAlchemy ORM queries.

.. warning::
    The PixelType values are not always properly translated by the
    `Rasterio <https://rasterio.readthedocs.io/en/stable/index.html>`_ library, so exporting a
    raster and re-importing it using this method will properly import the values but might not
    keep the same internal types.
"""
import struct
from sys import byteorder

import numpy as np
import pytest
import rasterio
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import text
from sqlalchemy.orm import declarative_base

from geoalchemy2 import Raster
from geoalchemy2 import RasterElement

# Tests imports
from tests import test_only_with_dialects

metadata = MetaData()
Base = declarative_base(metadata=metadata)


class Ocean(Base):  # type: ignore
    __tablename__ = "ocean"
    id = Column(Integer, primary_key=True)
    rast = Column(Raster)

    def __init__(self, rast):
        self.rast = rast


_DTYPE = {
    "?": [0, "?", 1],
    "u1": [2, "B", 1],
    "i1": [3, "b", 1],
    "B": [4, "B", 1],
    "i2": [5, "h", 2],
    "u2": [6, "H", 2],
    "i4": [7, "i", 4],
    "u4": [8, "I", 4],
    "f4": [10, "f", 4],
    "f8": [11, "d", 8],
}


def write_wkb_raster(dataset):
    """Creates a WKB raster from the given raster file with rasterio.
    :dataset: Rasterio dataset
    :returns: binary: Binary raster in WKB format

    This function was imported from
    https://github.com/nathancahill/wkb-raster/blob/master/wkb_raster.py
    and slightly adapted.
    """

    # Define format, see https://docs.python.org/3/library/struct.html
    format_string = "bHHddddddIHH"

    if byteorder == "big":
        endian = ">"
        endian_byte = 0
    elif byteorder == "little":
        endian = "<"
        endian_byte = 1

    # Write the raster header data.
    header = bytes()

    transform = dataset.transform.to_gdal()

    version = 0
    nBands = int(dataset.count)
    scaleX = transform[1]
    scaleY = transform[5]
    ipX = transform[0]
    ipY = transform[3]
    skewX = 0
    skewY = 0
    srid = int(dataset.crs.to_string().split("EPSG:")[1])
    width = int(dataset.meta.get("width"))
    height = int(dataset.meta.get("height"))

    if width > 65535 or height > 65535:
        raise ValueError("PostGIS does not support rasters with width or height greater than 65535")

    fmt = f"{endian}{format_string}"

    header = struct.pack(
        fmt,
        endian_byte,
        version,
        nBands,
        scaleX,
        scaleY,
        ipX,
        ipY,
        skewX,
        skewY,
        srid,
        width,
        height,
    )

    bands = []

    # Create band header data

    # not used - always False
    isOffline = False
    hasNodataValue = False

    if "nodata" in dataset.meta:
        hasNodataValue = True

    # not used - always False
    isNodataValue = False

    # unset
    reserved = False

    # # Based on the pixel type, determine the struct format, byte size and
    # # numpy dtype
    rasterio_dtype = dataset.meta.get("dtype")
    dt_short = np.dtype(rasterio_dtype).str[1:]
    pixtype, nodata_fmt, _ = _DTYPE[dt_short]

    # format binary -> :b
    binary_str = f"{isOffline:b}{hasNodataValue:b}{isNodataValue:b}{reserved:b}{pixtype:b}"
    # convert to int
    binary_decimal = int(binary_str, 2)

    # pack to 1 byte
    # 4 bits for ifOffline, hasNodataValue, isNodataValue, reserved
    # 4 bit for pixtype
    # -> 8 bit = 1 byte
    band_header = struct.pack("<b", binary_decimal)

    # Write the nodata value
    nodata = struct.pack(nodata_fmt, int(dataset.meta.get("nodata") or 0))

    for i in range(1, nBands + 1):
        band_array = dataset.read(i)

        # # Write the pixel values: width * height * size

        # numpy tobytes() method instead of packing with struct.pack()
        band_binary = band_array.reshape(width * height).tobytes()

        bands.append(band_header + nodata + band_binary)

    # join all bands
    allbands = bytes()
    for b in bands:
        allbands += b

    wkb = header + allbands

    return wkb


@test_only_with_dialects("postgresql")
class TestInsertRaster:
    @pytest.fixture(
        params=[
            "1BB",
            "2BUI",
            "4BUI",
            "8BSI",
            "8BUI",
            "16BSI",
            "16BUI",
            "32BSI",
            "32BUI",
            "32BF",
            "64BF",
        ]
    )
    def input_img(self, conn, tmpdir, request):
        """Create a TIFF image that will be imported as Raster."""
        pixel_type = request.param
        conn.execute(text("SET postgis.gdal_enabled_drivers = 'ENABLE_ALL';"))
        data = conn.execute(
            text(
                """SELECT
                    ST_AsTIFF(
                        ST_AsRaster(
                            ST_GeomFromText('POLYGON((0 0,1 1,0 1,0 0))'),
                            5,
                            6,
                            '{}'
                        ),
                        'GTiff'
                    );""".format(
                    pixel_type
                )
            )
        ).scalar()
        filename = tmpdir / "image.tiff"
        with open(filename, "wb") as f:
            f.write(data.tobytes())
        return filename, pixel_type

    def test_insert_raster(self, session, conn, input_img):
        """Insert a TIFF image into a raster column."""
        filename, pixel_type = input_img
        metadata.drop_all(conn, checkfirst=True)
        metadata.create_all(conn)

        # Load the image and transform it into a WKB
        with rasterio.open(str(filename), "r+") as dataset:
            dataset.crs = rasterio.crs.CRS.from_epsg(4326)
            expected_values = dataset.read()[0]
            raw_wkb = write_wkb_raster(dataset)

        # Insert a new raster element
        polygon_raster = RasterElement(raw_wkb)
        o = Ocean(polygon_raster)
        session.add(o)
        session.flush()

        # Check inserted values
        new_values = conn.execute(text("SELECT ST_DumpValues(rast, 1, false) FROM ocean;")).scalar()
        np.testing.assert_array_equal(
            np.array(new_values, dtype=expected_values.dtype), expected_values
        )



from __future__ import annotations

import binascii
import re
import struct
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Union

from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import functions
from sqlalchemy.sql.functions import FunctionElement
from sqlalchemy.types import to_instance

from geoalchemy2.exc import ArgumentError

BinasciiError = binascii.Error

function_registry: Set[str] = set()


class _SpatialElement:
    """The base class for public spatial elements.

    Args:
        data: The first argument passed to the constructor is the data wrapped
            by the ``_SpatialElement`` object being constructed.
        srid: An integer representing the spatial reference system. E.g. ``4326``.
            Default value is ``-1``, which means no/unknown reference system.
        extended: A boolean indicating whether the extended format (EWKT or EWKB)
            is used. Default is ``None``.

    """

    def __init__(self, data, srid: int = -1, extended: Optional[bool] = None) -> None:
        self.srid = srid
        self.data = data
        self.extended = extended

    def __str__(self) -> str:
        return self.desc

    def __repr__(self) -> str:
        return "<%s at 0x%x; %s>" % (
            self.__class__.__name__,
            id(self),
            self,
        )  # pragma: no cover

    def __eq__(self, other) -> bool:
        try:
            return (
                self.extended == other.extended
                and self.srid == other.srid
                and self.desc == other.desc
            )
        except AttributeError:
            return False

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.desc, self.srid, self.extended))

    def __getattr__(self, name):
        #
        # This is how things like lake.geom.ST_Buffer(2) creates
        # SQL expressions of this form:
        #
        # ST_Buffer(ST_GeomFromWKB(:ST_GeomFromWKB_1), :param_1)
        #

        # Raise an AttributeError when the attribute name doesn't start
        # with st_. This is to be nice with other libraries that use
        # some ducktyping (e.g. hasattr(element, "copy")) to determine
        # the type of the element.

        if name.lower() not in function_registry:
            raise AttributeError

        # We create our own _FunctionGenerator here, and use it in place of
        # SQLAlchemy's "func" object. This is to be able to "bind" the
        # function to the SQL expression. See also GenericFunction above.
        func_ = functions._FunctionGenerator(expr=self)
        return getattr(func_, name)

    def __getstate__(self) -> Dict[str, Any]:
        state = {
            "srid": self.srid,
            "data": str(self),
            "extended": self.extended,
        }
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self.srid = state["srid"]
        self.extended = state["extended"]
        self.data = self._data_from_desc(state["data"])

    @staticmethod
    def _data_from_desc(desc):
        raise NotImplementedError()  # pragma: no cover


class WKTElement(_SpatialElement):
    """Instances of this class wrap a WKT or EWKT value.

    Usage examples::

        wkt_element_1 = WKTElement('POINT(5 45)')
        wkt_element_2 = WKTElement('POINT(5 45)', srid=4326)
        wkt_element_3 = WKTElement('SRID=4326;POINT(5 45)', extended=True)
    """

    _REMOVE_SRID = re.compile("(SRID=([0-9]+); ?)?(.*)")

    geom_from: str = "ST_GeomFromText"
    geom_from_extended_version: str = "ST_GeomFromEWKT"

    def __init__(self, data: str, srid: int = -1, extended: Optional[bool] = None) -> None:
        if extended is None:
            extended = data.startswith("SRID=")
        if extended and srid == -1:
            # read srid from EWKT
            data_s = data.split(";")
            if len(data_s) != 2:
                raise ArgumentError("invalid EWKT string {}".format(data))
            header = data_s[0]
            try:
                srid = int(header[5:])
            except ValueError:
                raise ArgumentError("invalid EWKT string {}".format(data))
        _SpatialElement.__init__(self, data, srid, extended)

    @property
    def desc(self) -> str:
        """This element's description string."""
        return self.data

    @staticmethod
    def _data_from_desc(desc):
        return desc

    def as_wkt(self) -> WKTElement:
        if self.extended:
            srid_match = self._REMOVE_SRID.match(self.data)
            assert srid_match is not None
            return WKTElement(srid_match.group(3), self.srid, extended=False)
        return WKTElement(self.data, self.srid, self.extended)

    def as_ewkt(self) -> WKTElement:
        if not self.extended and self.srid != -1:
            data = f"SRID={self.srid};" + self.data
            return WKTElement(data, extended=True)
        return WKTElement(self.data, self.srid, self.extended)


class WKBElement(_SpatialElement):
    """Instances of this class wrap a WKB or EWKB value.

    Geometry values read from the database are converted to instances of this
    type. In most cases you won't need to create ``WKBElement`` instances
    yourself.

    If ``extended`` is ``True`` and ``srid`` is ``-1`` at construction time
    then the SRID will be read from the EWKB data.

    Note: you can create ``WKBElement`` objects from Shapely geometries
    using the :func:`geoalchemy2.shape.from_shape` function.
    """

    geom_from: str = "ST_GeomFromWKB"
    geom_from_extended_version: str = "ST_GeomFromEWKB"

    def __init__(
        self, data: Union[str, bytes, memoryview], srid: int = -1, extended: Optional[bool] = None
    ) -> None:
        if srid == -1 or extended is None or extended:
            # read srid from the EWKB
            #
            # WKB struct {
            #    byte    byteOrder;
            #    uint32  wkbType;
            #    uint32  SRID;
            #    struct  geometry;
            # }
            # byteOrder enum {
            #     WKB_XDR = 0,  // Most Significant Byte First
            #     WKB_NDR = 1,  // Least Significant Byte First
            # }
            # See https://trac.osgeo.org/postgis/browser/branches/3.0/doc/ZMSgeoms.txt
            # for more details about WKB/EWKB specifications.
            if isinstance(data, str):
                # SpatiaLite case
                # assume that the string is an hex value
                header = binascii.unhexlify(data[:18])
            else:
                header = data[:9]
            byte_order, wkb_type, wkb_srid = header[0], header[1:5], header[5:]
            byte_order_marker = "<I" if byte_order else ">I"
            wkb_type_int = (
                int(struct.unpack(byte_order_marker, wkb_type)[0]) if len(wkb_type) == 4 else 0
            )
            if extended is None:
                if not wkb_type_int:
                    extended = False
                else:
                    extended = extended or bool(wkb_type_int & 536870912)  # Check SRID bit
            if extended and srid == -1:
                wkb_srid = struct.unpack(byte_order_marker, wkb_srid)[0]
                srid = int(wkb_srid)
        _SpatialElement.__init__(self, data, srid, extended)

    @property
    def desc(self) -> str:
        """This element's description string."""
        if isinstance(self.data, str):
            # SpatiaLite case
            return self.data.lower()
        desc = str(binascii.hexlify(self.data), encoding="utf-8").lower()
        return desc

    @staticmethod
    def _data_from_desc(desc) -> bytes:
        desc = desc.encode(encoding="utf-8")
        return binascii.unhexlify(desc)

    def as_wkb(self) -> WKBElement:
        if self.extended:
            if isinstance(self.data, str):
                # SpatiaLite case
                # assume that the string is an hex value
                is_hex = True
                header = binascii.unhexlify(self.data[:10])
                byte_order, wkb_type = header[0], header[1:5]
            else:
                is_hex = False
                byte_order, wkb_type = self.data[0], self.data[1:5]

            byte_order_marker = "<I" if byte_order else ">I"
            wkb_type_int = (
                int(struct.unpack(byte_order_marker, wkb_type)[0]) if len(wkb_type) == 4 else 0
            )
            wkb_type_int &= 3758096383  # Set SRID bit to 0 and keep all other bits

            if is_hex:
                wkb_type_hex = binascii.hexlify(
                    wkb_type_int.to_bytes(4, "little" if byte_order else "big")
                )
                data = self.data[:2] + wkb_type_hex.decode("ascii") + self.data[18:]
            else:
                buffer = bytearray()
                buffer.extend(self.data[:1])
                buffer.extend(struct.pack(byte_order_marker, wkb_type_int))
                buffer.extend(self.data[9:])
                data = memoryview(buffer)
            return WKBElement(data, self.srid, extended=False)
        return WKBElement(self.data, self.srid)

    def as_ewkb(self) -> WKBElement:
        if not self.extended and self.srid != -1:
            if isinstance(self.data, str):
                # SpatiaLite case
                # assume that the string is an hex value
                header = binascii.unhexlify(self.data[:10])
                byte_order, wkb_type = header[0], header[1:5]
            else:
                byte_order, wkb_type = self.data[0], self.data[1:5]
            byte_order_marker = "<I" if byte_order else ">I"
            wkb_type_int = int(
                struct.unpack(byte_order_marker, wkb_type)[0] if len(wkb_type) == 4 else 0
            )
            wkb_type_int |= 536870912  # Set SRID bit to 1 and keep all other bits

            data: Union[str, memoryview]
            if isinstance(self.data, str):
                wkb_type_hex = binascii.hexlify(
                    wkb_type_int.to_bytes(4, "little" if byte_order else "big")
                )
                wkb_srid_hex = binascii.hexlify(
                    self.srid.to_bytes(4, "little" if byte_order else "big")
                )
                data = (
                    self.data[:2]
                    + wkb_type_hex.decode("ascii")
                    + wkb_srid_hex.decode("ascii")
                    + self.data[10:]
                )
            else:
                buffer = bytearray()
                buffer.extend(self.data[:1])
                buffer.extend(struct.pack(byte_order_marker, wkb_type_int))
                buffer.extend(struct.pack(byte_order_marker, self.srid))
                buffer.extend(self.data[5:])
                data = memoryview(buffer)

            return WKBElement(data, self.srid, extended=True)
        return WKBElement(self.data, self.srid)


class RasterElement(_SpatialElement):
    """Instances of this class wrap a ``raster`` value.

    Raster values read from the database are converted to instances of this type. In
    most cases you won't need to create ``RasterElement`` instances yourself.
    """

    geom_from_extended_version: str = "raster"

    def __init__(self, data: Union[str, bytes, memoryview]) -> None:
        # read srid from the WKB (binary or hexadecimal format)
        # The WKB structure is documented in the file
        # raster/doc/RFC2-WellKnownBinaryFormat of the PostGIS sources.
        bin_data: Union[str, bytes, memoryview]
        try:
            bin_data = binascii.unhexlify(data[:114])
        except BinasciiError:
            bin_data = data
            data = str(binascii.hexlify(data).decode(encoding="utf-8"))  # type: ignore
        byte_order = bin_data[0]
        srid = bin_data[53:57]
        srid = struct.unpack("<I" if byte_order else ">I", srid)[0]  # type: ignore
        _SpatialElement.__init__(self, data, int(srid), True)

    @property
    def desc(self) -> str:
        """This element's description string."""
        return self.data

    @staticmethod
    def _data_from_desc(desc):
        return desc


class CompositeElement(FunctionElement):
    """Instances of this class wrap a Postgres composite type."""

    inherit_cache: bool = False
    """The cache is disabled for this class."""

    def __init__(self, base, field, type_) -> None:
        self.name = field
        self.type = to_instance(type_)

        super(CompositeElement, self).__init__(base)


@compiles(CompositeElement)
def _compile_pgelem(expr, compiler, **kw) -> str:
    return "(%s).%s" % (compiler.process(expr.clauses, **kw), expr.name)


__all__: List[str] = [
    "_SpatialElement",
    "CompositeElement",
    "RasterElement",
    "WKBElement",
    "WKTElement",
]


def __dir__() -> List[str]:
    return __all__



import os
import platform
import re
import shutil
import sys

import pytest
from packaging import version
from pkg_resources import parse_version
from sqlalchemy import __version__ as SA_VERSION
from sqlalchemy import create_engine
from sqlalchemy import select as raw_select
from sqlalchemy import text
from sqlalchemy.event import listen
from sqlalchemy.exc import OperationalError
from sqlalchemy.sql import func

from geoalchemy2 import load_spatialite
from geoalchemy2 import load_spatialite_gpkg


class test_only_with_dialects:
    def __init__(self, *dialects):
        self.tested_dialects = dialects

    def __call__(self, test_obj):
        test_obj.tested_dialects = self.tested_dialects
        return test_obj


def get_postgis_major_version(bind):
    try:
        return version.parse(bind.execute(func.postgis_lib_version()).scalar()).major
    except OperationalError:
        return version.parse("0").major


def get_postgres_major_version(bind):
    try:
        return re.match(
            r"([0-9]*)\.([0-9]*).*",
            bind.execute(text("""SELECT current_setting('server_version');""")).scalar(),
        ).group(1)
    except OperationalError:
        return "0"


def skip_postgis1(bind):
    if get_postgis_major_version(bind) == 1:
        pytest.skip("requires PostGIS != 1")


def skip_postgis2(bind):
    if get_postgis_major_version(bind) == 2:
        pytest.skip("requires PostGIS != 2")


def skip_postgis3(bind):
    if get_postgis_major_version(bind) == 3:
        pytest.skip("requires PostGIS != 3")


def skip_case_insensitivity():
    return pytest.mark.skipif(
        parse_version(SA_VERSION) < parse_version("1.3.4"),
        reason="Case-insensitivity is only available for sqlalchemy>=1.3.4",
    )


def skip_pg12_sa1217(bind):
    if (
        parse_version(SA_VERSION) < parse_version("1.2.17")
        and int(get_postgres_major_version(bind)) >= 12
    ):
        pytest.skip("Reflection for PostgreSQL-12 is only supported by sqlalchemy>=1.2.17")


def skip_pypy(msg=None):
    if platform.python_implementation() == "PyPy":
        pytest.skip(msg if msg is not None else "Incompatible with PyPy")


def select(args):
    if version.parse(SA_VERSION) < version.parse("1.4"):
        return raw_select(args)
    else:
        return raw_select(*args)


def format_wkt(wkt):
    return wkt.replace(", ", ",")


def copy_and_connect_sqlite_db(input_db, tmp_db, engine_echo, dialect):
    if "SPATIALITE_LIBRARY_PATH" not in os.environ:
        pytest.skip("SPATIALITE_LIBRARY_PATH is not defined, skip SpatiaLite tests")

    shutil.copyfile(input_db, tmp_db)

    print("INPUT DB:", input_db)
    print("TEST DB:", tmp_db)

    db_url = f"{dialect}:///{tmp_db}"
    engine = create_engine(
        db_url, echo=engine_echo, execution_options={"schema_translate_map": {"gis": None}}
    )

    if dialect == "gpkg":
        listen(engine, "connect", load_spatialite_gpkg)
    else:
        listen(engine, "connect", load_spatialite)

    with engine.begin() as connection:
        print(
            "SPATIALITE VERSION:",
            connection.execute(text("SELECT spatialite_version();")).fetchone()[0],
        )
        print(
            "GEOS VERSION:",
            connection.execute(text("SELECT geos_version();")).fetchone()[0],
        )
        if sys.version_info.minor > 7:
            print(
                "PROJ VERSION:",
                connection.execute(text("SELECT proj_version();")).fetchone()[0],
            )
            print(
                "PROJ DB PATH:",
                connection.execute(text("SELECT PROJ_GetDatabasePath();")).fetchone()[0],
            )

    return engine


def check_indexes(conn, dialect_name, expected, table_name):
    """Check that actual indexes are equal to the expected ones."""
    index_query = {
        "postgresql": text(
            """SELECT indexname, indexdef
            FROM pg_indexes
            WHERE
                tablename = '{}'
            ORDER BY indexname;""".format(
                table_name
            )
        ),
        "sqlite": text(
            """SELECT *
            FROM geometry_columns
            WHERE f_table_name = '{}'
            ORDER BY f_table_name, f_geometry_column;""".format(
                table_name
            )
        ),
        "geopackage": text(
            """SELECT table_name, column_name, extension_name
            FROM gpkg_extensions
            WHERE table_name = '{}' and extension_name = 'gpkg_rtree_index'
            """.format(
                table_name
            )
        ),
    }

    # Query to check the indexes
    actual_indexes = conn.execute(index_query[dialect_name]).fetchall()

    expected_indexes = expected[dialect_name]
    if dialect_name == "postgresql":
        expected_indexes = [(i[0], re.sub("\n *", " ", i[1])) for i in expected_indexes]

    try:
        assert actual_indexes == expected_indexes
    except AssertionError as exc:
        print("###############################################")

        print("Expected indexes:")
        for i in expected_indexes:
            print(i)

        print("Actual indexes:")
        for i in actual_indexes:
            print(i)

        print("###############################################")

        raise exc



import re
from itertools import permutations

import pytest
from shapely import wkb
from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import func

from geoalchemy2.elements import CompositeElement
from geoalchemy2.elements import RasterElement
from geoalchemy2.elements import WKBElement
from geoalchemy2.elements import WKTElement
from geoalchemy2.exc import ArgumentError
from geoalchemy2.types import Geometry


@pytest.fixture
def geometry_table():
    table = Table("table", MetaData(), Column("geom", Geometry))
    return table


def eq_sql(a, b):
    a = re.sub(r"[\n\t]", "", str(a))
    assert a == b


class TestWKTElement:
    _srid = 3857  # expected srid
    _wkt = "POINT (1 2)"  # expected wkt
    _ewkt = "SRID=3857;POINT (1 2)"  # expected ewkt

    def test_ctor(self):
        e1 = WKTElement(self._wkt)
        e2 = WKTElement(self._wkt, extended=False)
        e3 = WKTElement(self._wkt, srid=self._srid)
        e4 = WKTElement(self._wkt, srid=self._srid, extended=True)
        e5 = WKTElement(self._wkt, srid=self._srid, extended=False)
        assert e1.desc == e2.desc == e3.desc == e4.desc == e5.desc == self._wkt
        assert e1.srid == e2.srid == -1
        assert e3.srid == e4.srid == e5.srid == self._srid
        assert e1.extended == e2.extended == e3.extended == (not e4.extended) == e5.extended

    def test_desc(self):
        e = WKTElement(self._wkt)
        assert e.desc == self._wkt

    def test_function_call(self):
        e = WKTElement(self._wkt)
        f = e.ST_Buffer(2)
        eq_sql(
            f,
            "ST_Buffer("
            "ST_GeomFromText(:ST_GeomFromText_1, :ST_GeomFromText_2), "
            ":ST_Buffer_1)",
        )
        assert f.compile().params == {
            "ST_Buffer_1": 2,
            "ST_GeomFromText_1": self._wkt,
            "ST_GeomFromText_2": -1,
        }

    def test_attribute_error(self):
        e = WKTElement(self._wkt)
        assert not hasattr(e, "foo")

    def test_pickle_unpickle(self):
        import pickle

        e = WKTElement(self._wkt, srid=3, extended=False)
        pickled = pickle.dumps(e)
        unpickled = pickle.loads(pickled)
        assert unpickled.srid == 3
        assert unpickled.extended is False
        assert unpickled.data == self._wkt
        f = unpickled.ST_Buffer(2)
        eq_sql(
            f,
            "ST_Buffer("
            "ST_GeomFromText(:ST_GeomFromText_1, :ST_GeomFromText_2), "
            ":ST_Buffer_1)",
        )
        assert f.compile().params == {
            "ST_Buffer_1": 2,
            "ST_GeomFromText_1": self._wkt,
            "ST_GeomFromText_2": 3,
        }

    def test_eq(self):
        a = WKTElement(self._wkt)
        b = WKTElement(self._wkt)
        assert a == b

    def test_hash(self):
        a = WKTElement(self._wkt)
        b = WKTElement("POINT(10 20)")
        c = WKTElement("POINT(10 20)")
        assert set([a, b, c]) == set([a, b, c])
        assert len(set([a, b, c])) == 2

    def test_as_wkt_as_ewkt(self):
        e1 = WKTElement(self._wkt)
        e2 = WKTElement(self._wkt, srid=self._srid)

        # No SRID
        e3 = e1.as_ewkt()
        assert e3.desc == self._wkt
        assert e3.srid == -1
        assert e3.extended is False

        # With SRID
        e4 = e2.as_ewkt()
        assert e4.desc == f"SRID={self._srid};{self._wkt}"
        assert e4.srid == self._srid
        assert e4.extended is True

        assert e3.as_wkt() == e3.as_wkt().as_wkt() == e1
        assert e4.as_wkt() == e4.as_wkt().as_wkt() == e2


class TestExtendedWKTElement:
    _srid = 3857  # expected srid
    _wkt = "POINT (1 2 3)"  # expected wkt
    _ewkt = "SRID=3857;POINT (1 2 3)"  # expected ewkt

    def test_ctor(self):
        arbitrary_srid = self._srid + 1
        e1 = WKTElement(self._ewkt)
        e2 = WKTElement(self._ewkt, extended=True)
        e3 = WKTElement(self._ewkt, srid=arbitrary_srid)
        e4 = WKTElement(self._ewkt, srid=arbitrary_srid, extended=True)
        e5 = WKTElement(self._ewkt, srid=arbitrary_srid, extended=False)
        assert e1.desc == e2.desc == e3.desc == e4.desc == e5.desc == self._ewkt
        assert e1.srid == e2.srid == self._srid
        assert e3.srid == e4.srid == e5.srid == arbitrary_srid
        assert e1.extended == e2.extended == e3.extended == e4.extended == (not e5.extended)

    def test_desc(self):
        e = WKTElement(self._ewkt)
        assert e.desc == self._ewkt

    def test_function_call(self):
        e = WKTElement(self._ewkt, extended=True)
        f = e.ST_Buffer(2)
        eq_sql(f, "ST_Buffer(" "ST_GeomFromEWKT(:ST_GeomFromEWKT_1), " ":ST_Buffer_1)")
        assert f.compile().params == {"ST_Buffer_1": 2, "ST_GeomFromEWKT_1": self._ewkt}

    def test_pickle_unpickle(self):
        import pickle

        e = WKTElement(self._ewkt, extended=True)
        pickled = pickle.dumps(e)
        unpickled = pickle.loads(pickled)
        assert unpickled.srid == self._srid
        assert unpickled.extended is True
        assert unpickled.data == self._ewkt
        f = unpickled.ST_Buffer(2)
        eq_sql(f, "ST_Buffer(" "ST_GeomFromEWKT(:ST_GeomFromEWKT_1), " ":ST_Buffer_1)")
        assert f.compile().params == {
            "ST_Buffer_1": 2,
            "ST_GeomFromEWKT_1": self._ewkt,
        }

    def test_unpack_srid_from_ewkt(self):
        """
        Unpack SRID from WKT struct (when it is not provided as arg)
        to ensure geometry result processor preserves query SRID.
        """
        e = WKTElement(self._ewkt, extended=True)
        assert e.srid == self._srid
        assert e.desc == self._ewkt

    def test_unpack_srid_from_ewkt_forcing_srid(self):
        e = WKTElement(self._ewkt, srid=9999, extended=True)
        assert e.srid == 9999
        assert e.desc == self._ewkt

    def test_unpack_srid_from_bad_ewkt(self):
        with pytest.raises(ArgumentError):
            WKTElement("SRID=BAD SRID;POINT (1 2 3)", extended=True)
        with pytest.raises(ArgumentError):
            WKTElement("SRID=BAD SRID;POINT (1 2 3)", extended=True)
        with pytest.raises(ArgumentError):
            WKTElement("SRID=1234;1234;1234;POINT (1 2 3)", extended=True)

    def test_eq(self):
        a = WKTElement(self._ewkt, extended=True)
        b = WKTElement(self._ewkt, extended=True)
        assert a == b

    def test_hash(self):
        a = WKTElement("SRID=3857;POINT (1 2 3)", extended=True)
        b = WKTElement("SRID=3857;POINT (10 20 30)", extended=True)
        c = WKTElement("SRID=3857;POINT (10 20 30)", extended=True)
        assert set([a, b, c]) == set([a, b, c])
        assert len(set([a, b, c])) == 2

    def test_missing_srid(self):
        with pytest.raises(ArgumentError, match="invalid EWKT string"):
            WKTElement(self._wkt, extended=True)

    def test_missing_semi_colon(self):
        with pytest.raises(ArgumentError, match="invalid EWKT string"):
            WKTElement("SRID=3857" + self._wkt, extended=True)

    def test_as_wkt_as_ewkt(self):
        arbitrary_srid = self._srid + 1
        e1 = WKTElement(self._ewkt)
        e2 = WKTElement(self._ewkt, srid=arbitrary_srid)

        # No arbitrary SRID
        e3 = e1.as_wkt()
        assert e3.desc == self._wkt
        assert e3.srid == self._srid
        assert e3.extended is False
        assert e3.as_ewkt() == e1

        # With arbitrary SRID
        e4 = e2.as_wkt()
        assert e4.desc == self._wkt
        assert e4.srid == arbitrary_srid
        assert e4.extended is False
        # The arbitrary SRID overwrites the original SRID in the EWKT string
        assert e4.as_ewkt() == WKTElement(f"SRID={arbitrary_srid};{self._wkt}")


class TestWKTElementFunction:
    def test_ST_Equal_WKTElement_WKTElement(self):
        expr = func.ST_Equals(WKTElement("POINT(1 2)"), WKTElement("POINT(1 2)"))
        eq_sql(
            expr,
            "ST_Equals("
            "ST_GeomFromText(:ST_GeomFromText_1, :ST_GeomFromText_2), "
            "ST_GeomFromText(:ST_GeomFromText_3, :ST_GeomFromText_4))",
        )
        assert expr.compile().params == {
            "ST_GeomFromText_1": "POINT(1 2)",
            "ST_GeomFromText_2": -1,
            "ST_GeomFromText_3": "POINT(1 2)",
            "ST_GeomFromText_4": -1,
        }

    def test_ST_Equal_Column_WKTElement(self, geometry_table):
        expr = func.ST_Equals(geometry_table.c.geom, WKTElement("POINT(1 2)"))
        eq_sql(
            expr,
            'ST_Equals("table".geom, ' "ST_GeomFromText(:ST_GeomFromText_1, :ST_GeomFromText_2))",
        )
        assert expr.compile().params == {
            "ST_GeomFromText_1": "POINT(1 2)",
            "ST_GeomFromText_2": -1,
        }


class TestExtendedWKTElementFunction:
    def test_ST_Equal_WKTElement_WKTElement(self):
        expr = func.ST_Equals(
            WKTElement("SRID=3857;POINT(1 2 3)", extended=True),
            WKTElement("SRID=3857;POINT(1 2 3)", extended=True),
        )
        eq_sql(
            expr,
            "ST_Equals("
            "ST_GeomFromEWKT(:ST_GeomFromEWKT_1), "
            "ST_GeomFromEWKT(:ST_GeomFromEWKT_2))",
        )
        assert expr.compile().params == {
            "ST_GeomFromEWKT_1": "SRID=3857;POINT(1 2 3)",
            "ST_GeomFromEWKT_2": "SRID=3857;POINT(1 2 3)",
        }

    def test_ST_Equal_Column_WKTElement(self, geometry_table):
        expr = func.ST_Equals(
            geometry_table.c.geom, WKTElement("SRID=3857;POINT(1 2 3)", extended=True)
        )
        eq_sql(expr, 'ST_Equals("table".geom, ' "ST_GeomFromEWKT(:ST_GeomFromEWKT_1))")
        assert expr.compile().params == {
            "ST_GeomFromEWKT_1": "SRID=3857;POINT(1 2 3)",
        }


class TestExtendedWKBElement:
    # _bin/_hex computed by following query:
    # SELECT ST_GeomFromEWKT('SRID=3;POINT(1 2)');
    _bin_ewkb = memoryview(
        b"\x01\x01\x00\x00 \x03\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00@"
    )
    _bin_wkb = memoryview(
        b"\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00@"
    )
    _hex_ewkb = str("010100002003000000000000000000f03f0000000000000040")
    _hex_wkb = str("0101000000000000000000f03f0000000000000040")
    _srid = 3  # expected srid
    _wkt = "POINT (1 2)"  # expected wkt

    def test_desc(self):
        e1 = WKBElement(self._bin_ewkb, extended=True)
        e2 = WKBElement(self._bin_ewkb)
        assert e1.desc == self._hex_ewkb
        assert e2.desc == self._hex_ewkb

    def test_desc_str(self):
        e = WKBElement(self._hex_ewkb)
        assert e.desc == self._hex_ewkb

    def test_function_call(self):
        e = WKBElement(self._bin_ewkb, extended=True)
        f = e.ST_Buffer(2)
        eq_sql(f, "ST_Buffer(" "ST_GeomFromEWKB(:ST_GeomFromEWKB_1), " ":ST_Buffer_1)")
        assert f.compile().params == {
            "ST_Buffer_1": 2,
            "ST_GeomFromEWKB_1": self._bin_ewkb,
        }

    def test_function_str(self):
        e = WKBElement(self._bin_ewkb, extended=True)
        assert isinstance(str(e), str)

    def test_pickle_unpickle(self):
        import pickle

        e = WKBElement(self._bin_ewkb, srid=self._srid, extended=True)
        pickled = pickle.dumps(e)
        unpickled = pickle.loads(pickled)
        assert unpickled.srid == self._srid
        assert unpickled.extended is True
        assert unpickled.data == bytes(self._bin_ewkb)
        f = unpickled.ST_Buffer(2)
        eq_sql(f, "ST_Buffer(" "ST_GeomFromEWKB(:ST_GeomFromEWKB_1), " ":ST_Buffer_1)")
        assert f.compile().params == {
            "ST_Buffer_1": 2,
            "ST_GeomFromEWKB_1": bytes(self._bin_ewkb),
        }

    def test_unpack_srid_from_bin(self):
        """
        Unpack SRID from WKB struct (when it is not provided as arg)
        to ensure geometry result processor preserves query SRID.
        """
        e = WKBElement(self._bin_ewkb, extended=True)
        assert e.srid == self._srid
        assert wkb.loads(bytes(e.data)).wkt == self._wkt

    def test_unpack_srid_from_bin_forcing_srid(self):
        e = WKBElement(self._bin_ewkb, srid=9999, extended=True)
        assert e.srid == 9999
        assert wkb.loads(bytes(e.data)).wkt == self._wkt

    def test_unpack_srid_from_hex(self):
        e = WKBElement(self._hex_ewkb, extended=True)
        assert e.srid == self._srid
        assert wkb.loads(e.data, hex=True).wkt == self._wkt

    def test_eq(self):
        a = WKBElement(self._bin_ewkb, extended=True)
        b = WKBElement(self._bin_ewkb, extended=True)
        assert a == b

    def test_hash(self):
        a = WKBElement(str("010100002003000000000000000000f03f0000000000000040"), extended=True)
        b = WKBElement(str("010100002003000000000000000000f02f0000000000000040"), extended=True)
        c = WKBElement(str("010100002003000000000000000000f02f0000000000000040"), extended=True)
        assert set([a, b, c]) == set([a, b, c])
        assert len(set([a, b, c])) == 2

    def test_as_wkt_as_ewkt(self):
        arbitrary_srid = self._srid + 1
        e1 = WKBElement(self._bin_ewkb)
        e2 = WKBElement(self._bin_ewkb, srid=arbitrary_srid)
        e3 = WKBElement(self._hex_ewkb)
        e4 = WKBElement(self._hex_ewkb, srid=arbitrary_srid)

        # Bin with no arbitrary SRID
        e5 = e1.as_wkb()
        assert e5.desc == self._hex_wkb
        assert e5.srid == self._srid
        assert e5.extended is False
        assert e5.as_ewkb() == e1

        # Bin with arbitrary SRID
        e6 = e2.as_wkb()
        assert e6.desc == self._hex_wkb
        assert e6.srid == arbitrary_srid
        assert e6.extended is False
        # The arbitrary SRID overwrites the original SRID in the EWKB string
        e6_ewkb = WKBElement(self._bin_ewkb, srid=arbitrary_srid)
        data = bytearray(e6_ewkb.data)
        data[5] = 4
        e6_ewkb.data = memoryview(data)
        assert e6.as_ewkb() == e6_ewkb

        # Hex with no arbitrary SRID
        e7 = e3.as_wkb()
        assert e7.desc == self._hex_wkb
        assert e7.srid == self._srid
        assert e7.extended is False
        assert e7.as_ewkb() == e3

        # Hex with arbitrary SRID
        e8 = e4.as_wkb()
        assert e8.desc == self._hex_wkb
        assert e8.srid == arbitrary_srid
        assert e8.extended is False
        # The arbitrary SRID overwrites the original SRID in the EWKB string
        e8_ewkb = WKBElement(self._hex_ewkb, srid=arbitrary_srid)
        e8_ewkb.data = e8_ewkb.data[:11] + "4" + e8_ewkb.data[12:]
        assert e8.as_ewkb() == e8_ewkb


class TestWKBElement:
    def test_desc(self):
        e = WKBElement(b"\x01\x02")
        assert e.desc == "0102"

    def test_function_call(self):
        e = WKBElement(b"\x01\x02")
        f = e.ST_Buffer(2)
        eq_sql(
            f,
            "ST_Buffer(" "ST_GeomFromWKB(:ST_GeomFromWKB_1, :ST_GeomFromWKB_2), " ":ST_Buffer_1)",
        )
        assert f.compile().params == {
            "ST_Buffer_1": 2,
            "ST_GeomFromWKB_1": b"\x01\x02",
            "ST_GeomFromWKB_2": -1,
        }

    def test_attribute_error(self):
        e = WKBElement(b"\x01\x02")
        assert not hasattr(e, "foo")

    def test_function_str(self):
        e = WKBElement(b"\x01\x02")
        assert isinstance(str(e), str)

    def test_eq(self):
        a = WKBElement(b"\x01\x02")
        b = WKBElement(b"\x01\x02")
        assert a == b

    def test_hash(self):
        a = WKBElement(b"\x01\x02")
        b = WKBElement(b"\x01\x03")
        c = WKBElement(b"\x01\x03")
        assert set([a, b, c]) == set([a, b, c])
        assert len(set([a, b, c])) == 2


class TestNotEqualSpatialElement:
    # _bin/_hex computed by following query:
    # SELECT ST_GeomFromEWKT('SRID=3;POINT(1 2)');
    _ewkb = memoryview(
        b"\x01\x01\x00\x00 \x03\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00@"
    )
    _wkb = wkb.loads(bytes(_ewkb)).wkb
    _hex = str("010100002003000000000000000000f03f0000000000000040")
    _srid = 3
    _wkt = "POINT (1 2)"
    _ewkt = "SRID=3;POINT (1 2)"

    def test_eq(self):
        a = WKBElement(self._ewkb, extended=True)
        b = WKBElement(self._wkb, srid=self._srid)
        c = WKTElement(self._wkt, srid=self._srid)
        d = WKTElement(self._ewkt, extended=True)
        e = WKBElement(self._hex, extended=True)
        assert a == a
        assert b == b
        assert c == c
        assert d == d
        assert e == e
        assert a == e and e == a

    def test_neq_other_types(self):
        a = WKBElement(self._ewkb, extended=True)
        b = WKBElement(self._wkb, srid=self._srid)
        c = WKTElement(self._wkt, srid=self._srid)
        d = WKTElement(self._ewkt, extended=True)
        e = WKBElement(self._hex, extended=True)
        all_elements = [a, b, c, d, None, 1, "test"]
        for i, j in permutations(all_elements, 2):
            assert i != j
        for i in all_elements[1:]:
            assert i != e and e != i


class TestRasterElement:
    rast_data = (
        b"\x01\x00\x00\x01\x00\x9a\x99\x99\x99\x99\x99\xc9?\x9a\x99\x99\x99\x99\x99"
        b"\xc9\xbf\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0?\x00"
        b"\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe6\x10\x00"
        b"\x00\x05\x00\x05\x00D\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x01\x01"
        b"\x01\x00\x00\x01\x01\x00\x00\x00\x01\x00\x00\x00\x00"
    )

    hex_rast_data = (
        "01000001009a9999999999c93f9a9999999999c9bf0000000000000000000000000000f03"
        "f00000000000000000000000000000000e610000005000500440001010101010101010100"
        "010101000001010000000100000000"
    )

    def test_desc(self):
        e = RasterElement(self.rast_data)
        assert e.desc == self.hex_rast_data
        assert e.srid == 4326
        e = RasterElement(self.hex_rast_data)
        assert e.desc == self.hex_rast_data
        assert e.srid == 4326

    def test_function_call(self):
        e = RasterElement(self.rast_data)
        f = e.ST_Height()
        eq_sql(f, "ST_Height(raster(:raster_1))")
        assert f.compile().params == {"raster_1": self.hex_rast_data}

    def test_pickle_unpickle(self):
        import pickle

        e = RasterElement(self.rast_data)
        assert e.srid == 4326
        assert e.extended is True
        assert e.data == self.hex_rast_data
        pickled = pickle.dumps(e)
        unpickled = pickle.loads(pickled)
        assert unpickled.srid == 4326
        assert unpickled.extended is True
        assert unpickled.data == self.hex_rast_data
        f = unpickled.ST_Height()
        eq_sql(f, "ST_Height(raster(:raster_1))")
        assert f.compile().params == {
            "raster_1": self.hex_rast_data,
        }

    def test_hash(self):
        new_hex_rast_data = self.hex_rast_data.replace("f", "e")
        a = WKBElement(self.hex_rast_data)
        b = WKBElement(new_hex_rast_data)
        c = WKBElement(new_hex_rast_data)
        assert set([a, b, c]) == set([a, b, c])
        assert len(set([a, b, c])) == 2


class TestCompositeElement:
    def test_compile(self):
        # text fixture
        metadata = MetaData()
        foo = Table("foo", metadata, Column("one", String))

        e = CompositeElement(foo.c.one, "geom", String)
        assert str(e) == "(foo.one).geom"



from textwrap import TextWrapper
from typing import Optional
from typing import Tuple
from typing import Union


def _wrap_docstring(docstring: str) -> str:
    wrapper = TextWrapper(width=100)
    lines = []
    for long_line in docstring.splitlines(keepends=False):
        lines.extend(wrapper.wrap(long_line))
    return "\n".join(lines)


def _get_docstring(name: str, doc: Union[None, str, Tuple[str, str]], type_: Optional[type]) -> str:
    doc_string_parts = []

    if isinstance(doc, tuple):
        doc_string_parts.append(_wrap_docstring(doc[0]))
        doc_string_parts.append("see https://postgis.net/docs/{0}.html".format(doc[1]))
    elif doc is not None:
        doc_string_parts.append(_wrap_docstring(doc))
        doc_string_parts.append("see https://postgis.net/docs/{0}.html".format(name))

    if type_ is not None:
        return_type_str = "{0}.{1}".format(type_.__module__, type_.__name__)
        doc_string_parts.append("Return type: :class:`{0}`.".format(return_type_str))

    return "\n\n".join(doc_string_parts)


def _replace_indent(text: str, indent: str) -> str:
    lines = []
    for i, line in enumerate(text.splitlines()):
        if i == 0 or not line.strip():
            lines.append(line)
        else:
            lines.append(f"{indent}{line}")
    return "\n".join(lines)


def _generate_stubs() -> str:
    """Generates type stubs for the dynamic functions described in `geoalchemy2/_functions.py`."""
    from geoalchemy2._functions import _FUNCTIONS
    from geoalchemy2.functions import ST_AsGeoJSON

    header = '''\
# this file is automatically generated
from typing import List

from sqlalchemy.sql import functions
from sqlalchemy.sql.elements import ColumnElement

import geoalchemy2.types

class GenericFunction(functions.GenericFunction): ...

class TableRowElement(ColumnElement):
    inherit_cache: bool = ...
    """The cache is disabled for this class."""

    def __init__(self, selectable: bool) -> None: ...
    @property
    def _from_objects(self) -> List[bool]: ...  # type: ignore[override]
'''
    stub_file_parts = [header]

    functions = _FUNCTIONS.copy()
    functions.insert(0, ("ST_AsGeoJSON", None, ST_AsGeoJSON.__doc__))

    for name, type_, doc_parts in functions:
        doc = _replace_indent(_get_docstring(name, doc_parts, type_), "    ")

        if type_ is None:
            type_str = ""
        else:
            type_str = f"\n\n    type = {type_.__module__}.{type_.__name__}()"

        signature = f'''\
class {name}(GenericFunction):
    """
    {doc}
    """{type_str}
'''
        stub_file_parts.append(signature)

    return "\n".join(stub_file_parts)



"""This module defines specific functions for Postgresql dialect."""
from sqlalchemy import Index
from sqlalchemy import text
from sqlalchemy.sql import func
from sqlalchemy.sql import select

from geoalchemy2.admin.dialects.common import _check_spatial_type
from geoalchemy2.admin.dialects.common import _format_select_args
from geoalchemy2.admin.dialects.common import _spatial_idx_name
from geoalchemy2.admin.dialects.common import setup_create_drop
from geoalchemy2.types import Geography
from geoalchemy2.types import Geometry


def check_management(column):
    """Check if the column should be managed."""
    return getattr(column.type, "use_typmod", None) is False


def create_spatial_index(bind, table, col):
    """Create spatial index on the given column."""
    if col.type.use_N_D_index:
        postgresql_ops = {col.name: "gist_geometry_ops_nd"}
    else:
        postgresql_ops = {}
    idx = Index(
        _spatial_idx_name(table.name, col.name),
        col,
        postgresql_using="gist",
        postgresql_ops=postgresql_ops,
        _column_flag=True,
    )
    idx.create(bind=bind)


def reflect_geometry_column(inspector, table, column_info):
    """Reflect a column of type Geometry with Postgresql dialect."""
    if not isinstance(column_info.get("type"), Geometry):
        return
    geo_type = column_info["type"]
    geometry_type = geo_type.geometry_type
    coord_dimension = geo_type.dimension
    if geometry_type.endswith("ZM"):
        coord_dimension = 4
    elif geometry_type[-1] in ["Z", "M"]:
        coord_dimension = 3

    # Query to check a given column has spatial index
    if table.schema is not None:
        schema_part = " AND nspname = '{}'".format(table.schema)
    else:
        schema_part = ""

    has_index_query = """SELECT (indexrelid IS NOT NULL) AS has_index
        FROM (
            SELECT
                    n.nspname,
                    c.relname,
                    c.oid AS relid,
                    a.attname,
                    a.attnum
            FROM pg_attribute a
            INNER JOIN pg_class c ON (a.attrelid=c.oid)
            INNER JOIN pg_type t ON (a.atttypid=t.oid)
            INNER JOIN pg_namespace n ON (c.relnamespace=n.oid)
            WHERE t.typname='geometry'
                    AND c.relkind='r'
        ) g
        LEFT JOIN pg_index i ON (g.relid = i.indrelid AND g.attnum = ANY(i.indkey))
        WHERE relname = '{}' AND attname = '{}'{};
    """.format(
        table.name, column_info["name"], schema_part
    )
    spatial_index = inspector.bind.execute(text(has_index_query)).scalar()

    # Set attributes
    column_info["type"].geometry_type = geometry_type
    column_info["type"].dimension = coord_dimension
    column_info["type"].spatial_index = bool(spatial_index)

    # Spatial indexes are automatically reflected with PostgreSQL dialect
    column_info["type"]._spatial_index_reflected = True


def before_create(table, bind, **kw):
    """Handle spatial indexes during the before_create event."""
    dialect, gis_cols, regular_cols = setup_create_drop(table, bind, check_management)

    # Remove the spatial indexes from the table metadata because they should not be
    # created during the table.create() step since the associated columns do not exist
    # at this time.
    table.info["_after_create_indexes"] = []
    current_indexes = set(table.indexes)
    for idx in current_indexes:
        for col in table.info["_saved_columns"]:
            if (
                _check_spatial_type(col.type, Geometry, dialect) and check_management(col)
            ) and col in idx.columns.values():
                table.indexes.remove(idx)
                if idx.name != _spatial_idx_name(table.name, col.name) or not getattr(
                    col.type, "spatial_index", False
                ):
                    table.info["_after_create_indexes"].append(idx)


def after_create(table, bind, **kw):
    """Handle spatial indexes during the after_create event."""
    # Restore original column list including managed Geometry columns
    dialect = bind.dialect

    table.columns = table.info.pop("_saved_columns")

    for col in table.columns:
        # Add the managed Geometry columns with AddGeometryColumn()
        if _check_spatial_type(col.type, Geometry, dialect) and check_management(col):
            dimension = col.type.dimension
            args = [table.schema] if table.schema else []
            args.extend([table.name, col.name, col.type.srid, col.type.geometry_type, dimension])
            if col.type.use_typmod is not None:
                args.append(col.type.use_typmod)

            stmt = select(*_format_select_args(func.AddGeometryColumn(*args)))
            stmt = stmt.execution_options(autocommit=True)
            bind.execute(stmt)

        # Add spatial indices for the Geometry and Geography columns
        if (
            _check_spatial_type(col.type, (Geometry, Geography), dialect)
            and col.type.spatial_index is True
        ):
            # If the index does not exist, define it and create it
            if not [i for i in table.indexes if col in i.columns.values()] and check_management(
                col
            ):
                create_spatial_index(bind, table, col)

    for idx in table.info.pop("_after_create_indexes"):
        table.indexes.add(idx)
        idx.create(bind=bind)


def before_drop(table, bind, **kw):
    """Handle spatial indexes during the before_drop event."""
    dialect, gis_cols, regular_cols = setup_create_drop(table, bind, check_management)

    # Drop the managed Geometry columns
    for col in gis_cols:
        args = [table.schema] if table.schema else []
        args.extend([table.name, col.name])

        stmt = select(*_format_select_args(func.DropGeometryColumn(*args)))
        stmt = stmt.execution_options(autocommit=True)
        bind.execute(stmt)


def after_drop(table, bind, **kw):
    """Handle spatial indexes during the after_drop event."""
    # Restore original column list including managed Geometry columns
    saved_cols = table.info.pop("_saved_columns", None)
    if saved_cols is not None:
        table.columns = saved_cols



import os
from pathlib import Path

import pytest
from sqlalchemy import MetaData
from sqlalchemy import create_engine
from sqlalchemy.dialects.mysql.base import MySQLDialect
from sqlalchemy.dialects.sqlite.base import SQLiteDialect
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import sessionmaker

from geoalchemy2.alembic_helpers import _monkey_patch_get_indexes_for_mysql
from geoalchemy2.alembic_helpers import _monkey_patch_get_indexes_for_sqlite

from . import copy_and_connect_sqlite_db
from . import get_postgis_major_version
from . import get_postgres_major_version
from .schema_fixtures import *  # noqa


def pytest_addoption(parser):
    parser.addoption(
        "--postgresql_dburl",
        action="store",
        help="PostgreSQL DB URL used for tests (`postgresql://user:password@host:port/dbname`).",
    )
    parser.addoption(
        "--sqlite_spatialite3_dburl",
        action="store",
        help="SQLite DB URL used for tests with SpatiaLite3 (`sqlite:///path_to_db_file`).",
    )
    parser.addoption(
        "--sqlite_spatialite4_dburl",
        action="store",
        help="SQLite DB URL used for tests with SpatiaLite4 (`sqlite:///path_to_db_file`).",
    )
    parser.addoption(
        "--sqlite_geopackage_dburl",
        action="store",
        help="SQLite DB URL used for tests with GeoPackage (`gpkg:///path_to_db_file.gpkg`).",
    )
    parser.addoption(
        "--mysql_dburl",
        action="store",
        help="MySQL DB URL used for tests with MySQL (`mysql://user:password@host:port/dbname`).",
    )
    parser.addoption(
        "--engine-echo",
        action="store_true",
        default=False,
        help="If set to True, all statements of the engine are logged.",
    )


def pytest_generate_tests(metafunc):
    if "db_url" in metafunc.fixturenames:
        sqlite_dialects = ["sqlite-spatialite3", "sqlite-spatialite4", "geopackage"]
        dialects = None

        if metafunc.module.__name__ == "tests.test_functional_postgresql":
            dialects = ["postgresql"]
        elif metafunc.module.__name__ == "tests.test_functional_sqlite":
            dialects = sqlite_dialects
        elif metafunc.module.__name__ == "tests.test_functional_mysql":
            dialects = ["mysql"]
        elif metafunc.module.__name__ == "tests.test_functional_geopackage":
            dialects = ["geopackage"]

        if getattr(metafunc.function, "tested_dialects", False):
            dialects = metafunc.function.tested_dialects
        elif getattr(metafunc.cls, "tested_dialects", False):
            dialects = metafunc.cls.tested_dialects

        if dialects is None:
            dialects = ["mysql", "postgresql"] + sqlite_dialects

        if "sqlite" in dialects:
            # Order dialects
            dialects = [i for i in dialects if i != "sqlite"] + sqlite_dialects

        metafunc.parametrize("db_url", dialects, indirect=True)


@pytest.fixture(scope="session")
def db_url_postgresql(request):
    return (
        request.config.getoption("--postgresql_dburl")
        or os.getenv("PYTEST_POSTGRESQL_DB_URL")
        or "postgresql://gis:gis@localhost/gis"
    )


@pytest.fixture(scope="session")
def db_url_mysql(request, tmpdir_factory):
    return (
        request.config.getoption("--mysql_dburl")
        or os.getenv("PYTEST_MYSQL_DB_URL")
        or "mysql://gis:gis@localhost/gis"
    )


@pytest.fixture(scope="session")
def db_url_sqlite_spatialite3(request, tmpdir_factory):
    return (
        request.config.getoption("--sqlite_spatialite3_dburl")
        or os.getenv("PYTEST_SPATIALITE3_DB_URL")
        or f"sqlite:///{Path(__file__).parent / 'data' / 'spatialite_lt_4.sqlite'}"
    )


@pytest.fixture(scope="session")
def db_url_sqlite_spatialite4(request, tmpdir_factory):
    return (
        request.config.getoption("--sqlite_spatialite4_dburl")
        or os.getenv("PYTEST_SPATIALITE4_DB_URL")
        or f"sqlite:///{Path(__file__).parent / 'data' / 'spatialite_ge_4.sqlite'}"
    )


@pytest.fixture(scope="session")
def db_url_geopackage(request, tmpdir_factory):
    return (
        request.config.getoption("--sqlite_geopackage_dburl")
        or os.getenv("PYTEST_GEOPACKAGE_DB_URL")
        or f"gpkg:///{Path(__file__).parent / 'data' / 'spatialite_geopackage.gpkg'}"
    )


@pytest.fixture(scope="session")
def db_url(
    request,
    db_url_postgresql,
    db_url_sqlite_spatialite3,
    db_url_sqlite_spatialite4,
    db_url_geopackage,
    db_url_mysql,
):
    if request.param == "postgresql":
        return db_url_postgresql
    if request.param == "mysql":
        return db_url_mysql
    elif request.param == "sqlite-spatialite3":
        return db_url_sqlite_spatialite3
    elif request.param == "sqlite-spatialite4":
        return db_url_sqlite_spatialite4
    elif request.param == "geopackage":
        return db_url_geopackage
    return None


@pytest.fixture(scope="session")
def _engine_echo(request):
    _engine_echo = request.config.getoption("--engine-echo")
    return _engine_echo


@pytest.fixture
def engine(tmpdir, db_url, _engine_echo):
    """Provide an engine to test database."""
    if db_url.startswith("sqlite:///"):
        # Copy the input SQLite DB to a temporary file and return an engine to it
        input_url = str(db_url)[10:]
        output_file = "test_spatial_db.sqlite"
        return copy_and_connect_sqlite_db(input_url, tmpdir / output_file, _engine_echo, "sqlite")

    if db_url.startswith("gpkg:///"):
        # Copy the input SQLite DB to a temporary file and return an engine to it
        input_url = str(db_url)[8:]
        output_file = "test_spatial_db.gpkg"
        return copy_and_connect_sqlite_db(input_url, tmpdir / output_file, _engine_echo, "gpkg")

    # For other dialects the engine is directly returned
    engine = create_engine(db_url, echo=_engine_echo)
    engine.update_execution_options(search_path=["gis", "public"])
    return engine


@pytest.fixture
def check_spatialite():
    if "SPATIALITE_LIBRARY_PATH" not in os.environ:
        pytest.skip("SPATIALITE_LIBRARY_PATH is not defined, skip SpatiaLite tests")


@pytest.fixture
def dialect_name(engine):
    return engine.dialect.name


@pytest.fixture
def conn(engine):
    """Provide a connection to test database."""
    with engine.connect() as connection:
        trans = connection.begin()
        yield connection
        trans.rollback()


@pytest.fixture
def session(engine, conn):
    Session = sessionmaker(bind=conn)
    with Session(bind=conn) as session:
        yield session


@pytest.fixture
def schema(engine):
    if engine.dialect.name == "postgresql":
        return "gis"
    else:
        return None


@pytest.fixture
def metadata():
    return MetaData()


@pytest.fixture()
def base(metadata):
    return declarative_base(metadata=metadata)


@pytest.fixture
def postgis_version(conn):
    return get_postgis_major_version(conn)


@pytest.fixture
def postgres_major_version(conn):
    return get_postgres_major_version(conn)


@pytest.fixture(autouse=True)
def reset_alembic_monkeypatch():
    """Disable Alembic monkeypatching by default."""
    try:
        normal_behavior_sqlite = SQLiteDialect._get_indexes_normal_behavior
        SQLiteDialect.get_indexes = normal_behavior_sqlite
        SQLiteDialect._get_indexes_normal_behavior = normal_behavior_sqlite

        normal_behavior_mysql = MySQLDialect._get_indexes_normal_behavior
        MySQLDialect.get_indexes = normal_behavior_mysql
        MySQLDialect._get_indexes_normal_behavior = normal_behavior_mysql
    except AttributeError:
        pass


@pytest.fixture()
def use_alembic_monkeypatch():
    """Enable Alembic monkeypatching ."""
    _monkey_patch_get_indexes_for_sqlite()
    _monkey_patch_get_indexes_for_mysql()


@pytest.fixture
def setup_tables(conn, metadata):
    metadata.drop_all(conn, checkfirst=True)
    metadata.create_all(conn)
    yield



import os

import pytest
from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import create_engine
from sqlalchemy import text
from sqlalchemy.event import listen

from geoalchemy2 import Geometry
from geoalchemy2 import load_spatialite_gpkg

from .schema_fixtures import TransformedGeometry


class TestAdmin:
    def test_create_gpkg(self, tmpdir, _engine_echo, check_spatialite):
        """Test GeoPackage creation."""
        # Create empty GeoPackage
        tmp_db = tmpdir / "test_spatial_db.gpkg"
        db_url = f"gpkg:///{tmp_db}"
        engine = create_engine(
            db_url, echo=_engine_echo, execution_options={"schema_translate_map": {"gis": None}}
        )

        # Check that the DB is empty
        raw_conn = engine.connect()
        assert not raw_conn.execute(
            text("PRAGMA main.table_info('gpkg_geometry_columns');")
        ).fetchall()
        raw_conn.connection.dbapi_connection.enable_load_extension(True)
        raw_conn.connection.dbapi_connection.load_extension(os.getenv("SPATIALITE_LIBRARY_PATH"))
        raw_conn.connection.dbapi_connection.enable_load_extension(False)
        assert not raw_conn.execute(
            text("PRAGMA main.table_info('gpkg_geometry_columns');")
        ).fetchall()
        assert raw_conn.execute(text("SELECT HasGeoPackage();")).scalar()
        assert not raw_conn.execute(text("SELECT CheckGeoPackageMetaData();")).scalar()

        # Check that the DB is properly initialized using load_spatialite_gpkg
        listen(engine, "connect", load_spatialite_gpkg)
        conn = engine.connect()
        assert conn.execute(text("SELECT HasGeoPackage();")).scalar()
        assert conn.execute(text("SELECT CheckGeoPackageMetaData();")).scalar()

        # Create a new table in this manually created GeoPackage
        t = Table(
            "a_table",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("geom", Geometry(srid=2154)),
        )
        t.create(conn)
        t.drop(conn)

    def test_manual_initialization(self, tmpdir, _engine_echo, check_spatialite):
        """Test GeoPackage creation."""
        # Create empty GeoPackage
        tmp_db = tmpdir / "test_spatial_db.gpkg"
        db_url = f"gpkg:///{tmp_db}"
        engine = create_engine(
            db_url, echo=_engine_echo, execution_options={"schema_translate_map": {"gis": None}}
        )

        # Check that the DB is empty
        raw_conn = engine.connect()
        assert not raw_conn.execute(
            text("PRAGMA main.table_info('gpkg_geometry_columns');")
        ).fetchall()
        raw_conn.connection.dbapi_connection.enable_load_extension(True)
        raw_conn.connection.dbapi_connection.load_extension(os.environ["SPATIALITE_LIBRARY_PATH"])
        raw_conn.connection.dbapi_connection.enable_load_extension(False)
        assert not raw_conn.execute(
            text("PRAGMA main.table_info('gpkg_geometry_columns');")
        ).fetchall()
        assert raw_conn.execute(text("SELECT HasGeoPackage();")).scalar()
        assert not raw_conn.execute(text("SELECT CheckGeoPackageMetaData();")).scalar()

        # Create a new table in this manually created GeoPackage
        t = Table(
            "a_table",
            MetaData(),
            Column("id", Integer, primary_key=True),
        )
        t.create(raw_conn)

        # Check that the table was properly created
        res = raw_conn.execute(text("PRAGMA main.table_info(a_table)")).fetchall()
        assert res == [
            (0, "id", "INTEGER", 1, None, 1),
        ]

        t.drop(raw_conn)

        # Manual initialization
        raw_conn.execute(text("SELECT gpkgCreateBaseTables();"))

        # Create again a new table in this manually created GeoPackage
        t = Table(
            "a_table",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("geom", Geometry(srid=2154)),
        )
        t.create(raw_conn)

        # Check that the table and the spatial indexes were properly created
        res = raw_conn.execute(text("PRAGMA main.table_info(a_table)")).fetchall()
        assert res == [
            (0, "id", "INTEGER", 1, None, 1),
            (1, "geom", "GEOMETRY", 0, None, 0),
        ]
        res = raw_conn.execute(text("PRAGMA main.table_info(rtree_a_table_geom)")).fetchall()
        assert len(res) > 0

        t.drop(raw_conn)

        # Check that the table and the spatial indexes were properly removed
        res = raw_conn.execute(text("PRAGMA main.table_info(a_table)")).fetchall()
        assert res == []
        res = raw_conn.execute(text("PRAGMA main.table_info(rtree_a_table_geom)")).fetchall()
        assert res == []

        # Check that the DB is properly initialized
        assert raw_conn.execute(text("SELECT HasGeoPackage();")).scalar()
        assert raw_conn.execute(text("SELECT CheckGeoPackageMetaData();")).scalar()

    def test_multi_geom_cols_fail(self, conn):
        t = Table(
            "a_table",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("geom", Geometry()),
            Column("geom_2", Geometry()),
        )
        with pytest.raises(
            ValueError,
            match=r"Only one geometry column is allowed for a table stored in a GeoPackage\.",
        ):
            t.create(conn)

    def test_add_srid(self, conn):
        t = Table(
            "a_table",
            MetaData(),
            Column("id", Integer, primary_key=True),
            Column("geom", Geometry(srid=3857)),
        )
        check_srid_query = text("""SELECT srs_id FROM gpkg_spatial_ref_sys WHERE srs_id = 3857;""")
        assert not conn.execute(check_srid_query).fetchall()
        t.create(conn)
        assert conn.execute(check_srid_query).fetchall()


class TestIndex:
    def test_index_gpkg(self, conn, Lake, setup_tables):
        assert (
            conn.execute(
                text(
                    """SELECT COUNT(*) FROM gpkg_extensions
                WHERE table_name = 'lake'
                    AND column_name = 'geom'
                    AND extension_name = 'gpkg_rtree_index';"""
                )
            ).scalar()
            == 1
        )

    def test_type_decorator_index_gpkg(self, conn, base, metadata):
        class LocalPoint(base):
            __tablename__ = "local_point"
            id = Column(Integer, primary_key=True)
            geom = Column(TransformedGeometry(db_srid=2154, app_srid=4326, geometry_type="POINT"))

        metadata.drop_all(conn, checkfirst=True)
        metadata.create_all(conn)
        assert (
            conn.execute(
                text(
                    """SELECT COUNT(*) FROM gpkg_extensions
                WHERE table_name = 'local_point'
                    AND column_name = 'geom'
                    AND extension_name = 'gpkg_rtree_index';"""
                )
            ).scalar()
            == 1
        )


class TestMiscellaneous:
    def test_load_spatialite_gpkg(self, tmpdir, _engine_echo, check_spatialite):
        # Create empty DB
        tmp_db = tmpdir / "test_spatial_db.sqlite"
        db_url = f"sqlite:///{tmp_db}"
        engine = create_engine(
            db_url, echo=_engine_echo, execution_options={"schema_translate_map": {"gis": None}}
        )
        conn = engine.connect()

        assert not conn.execute(text("PRAGMA main.table_info('gpkg_geometry_columns')")).fetchall()
        assert not conn.execute(text("PRAGMA main.table_info('gpkg_spatial_ref_sys')")).fetchall()

        load_spatialite_gpkg(conn.connection.dbapi_connection, None)

        assert conn.execute(text("SELECT CheckGeoPackageMetaData();")).scalar() == 1
        assert conn.execute(text("PRAGMA main.table_info('gpkg_geometry_columns')")).fetchall()
        assert conn.execute(text("PRAGMA main.table_info('gpkg_spatial_ref_sys')")).fetchall()

        # Check that spatial_ref_sys table was properly populated
        nb_srid = conn.execute(text("SELECT COUNT(*) FROM gpkg_spatial_ref_sys;")).scalar()
        assert nb_srid == 3

    def test_load_spatialite_no_env_variable(self, monkeypatch, conn):
        monkeypatch.delenv("SPATIALITE_LIBRARY_PATH")
        with pytest.raises(RuntimeError):
            load_spatialite_gpkg(conn.connection.dbapi_connection, None)



import re
from pathlib import Path

from sqlalchemy import select
from sqlalchemy.sql import func

import geoalchemy2.functions
from geoalchemy2._functions_helpers import _generate_stubs
from geoalchemy2.types import Raster  # NOQA

#
# Importing geoalchemy2 actually registers the GeoAlchemy generic
# functions in SQLAlchemy's function registry.
#


def eq_sql(a, b):
    a = re.sub(r"[\n\t]", "", str(a))
    assert a == b


def _test_simple_func(name):
    eq_sql(
        getattr(func, name)(1).select(),
        'SELECT %(name)s(:%(name)s_2) AS "%(name)s_1"' % dict(name=name),
    )


def _test_geometry_returning_func(name):
    eq_sql(
        getattr(func, name)(1).select(),
        'SELECT ST_AsEWKB(%(name)s(:%(name)s_2)) AS "%(name)s_1"' % dict(name=name),
    )


def _test_geography_returning_func(name):
    eq_sql(
        getattr(func, name)(1).select(),
        'SELECT ST_AsBinary(%(name)s(:%(name)s_2)) AS "%(name)s_1"' % dict(name=name),
    )


def _test_raster_returning_func(name, *args, **kwargs):
    eq_sql(
        getattr(func, name)(1, *args, **kwargs).select(),
        'SELECT raster(%(name)s(:%(name)s_2)) AS "%(name)s_1"' % dict(name=name),
    )


def test_stubs_up_to_date() -> None:
    geoalchemy2_path = Path(geoalchemy2.__file__).parent
    current_stubs = (geoalchemy2_path / "functions.pyi").read_text()
    generated_stubs = _generate_stubs()
    assert current_stubs == generated_stubs


def test_predefined_function() -> None:
    geom = geoalchemy2.functions.ST_GeomFromText("POINT(0, 0)")
    eq_sql(
        select(geoalchemy2.functions.ST_Buffer(geom, 1.0)),
        "SELECT ST_AsEWKB(ST_Buffer(ST_GeomFromText(:ST_GeomFromText_1), "
        ':ST_Buffer_2)) AS "ST_Buffer_1"',
    )


#
# Geometry Constructors
#
def test_ST_Collect():
    _test_geometry_returning_func("ST_Collect")


def test_ST_BdPolyFromText():
    _test_geometry_returning_func("ST_BdPolyFromText")


def test_ST_BdMPolyFromText():
    _test_geometry_returning_func("ST_BdMPolyFromText")


def test_ST_Box2dFromGeoHash():
    _test_simple_func("ST_Box2dFromGeoHash")


def test_ST_GeogFromText():
    _test_geography_returning_func("ST_GeogFromText")


def test_ST_GeographyFromText():
    _test_geography_returning_func("ST_GeographyFromText")


def test_ST_GeogFromWKB():
    _test_geography_returning_func("ST_GeogFromWKB")


def test_ST_GeomFromTWKB():
    _test_geometry_returning_func("ST_GeomFromTWKB")


def test_ST_GeomCollFromText():
    _test_geometry_returning_func("ST_GeomCollFromText")


def test_ST_GeomFromEWKB():
    _test_geometry_returning_func("ST_GeomFromEWKB")


def test_ST_GeomFromEWKT():
    _test_geometry_returning_func("ST_GeomFromEWKT")


def test_ST_GeometryFromText():
    _test_geometry_returning_func("ST_GeometryFromText")


def test_ST_GeomFromGeoHash():
    _test_geometry_returning_func("ST_GeomFromGeoHash")


def test_ST_GeomFromGML():
    _test_geometry_returning_func("ST_GeomFromGML")


def test_ST_GeomFromGeoJSON():
    _test_geometry_returning_func("ST_GeomFromGeoJSON")


def test_ST_GeomFromKML():
    _test_geometry_returning_func("ST_GeomFromKML")


def test_ST_GMLToSQL():
    _test_geometry_returning_func("ST_GMLToSQL")


def test_ST_GeomFromText():
    _test_geometry_returning_func("ST_GeomFromText")


def test_ST_GeomFromWKB():
    _test_geometry_returning_func("ST_GeomFromWKB")


def test_ST_LineFromEncodedPolyline():
    _test_geometry_returning_func("ST_LineFromEncodedPolyline")


def test_ST_LineFromMultiPoint():
    _test_geometry_returning_func("ST_LineFromMultiPoint")


def test_ST_LineFromText():
    _test_geometry_returning_func("ST_LineFromText")


def test_ST_LineFromWKB():
    _test_geometry_returning_func("ST_LineFromWKB")


def test_ST_LinestringFromWKB():
    _test_geometry_returning_func("ST_LinestringFromWKB")


def test_ST_MakeBox2D():
    _test_simple_func("ST_MakeBox2D")


def test_ST_3DMakeBox():
    _test_simple_func("ST_3DMakeBox")


def test_ST_MakeLine():
    _test_geometry_returning_func("ST_MakeLine")


def test_ST_MakeEnvelope():
    _test_geometry_returning_func("ST_MakeEnvelope")


def test_ST_MakePolygon():
    _test_geometry_returning_func("ST_MakePolygon")


def test_ST_MakePoint():
    _test_geometry_returning_func("ST_MakePoint")


def test_ST_MakePointM():
    _test_geometry_returning_func("ST_MakePointM")


def test_ST_MLineFromText():
    _test_geometry_returning_func("ST_MLineFromText")


def test_ST_MPointFromText():
    _test_geometry_returning_func("ST_MPointFromText")


def test_ST_MPolyFromText():
    _test_geometry_returning_func("ST_MPolyFromText")


def test_ST_Point():
    _test_geometry_returning_func("ST_Point")


def test_ST_PointFromGeoHash():
    _test_geometry_returning_func("ST_PointFromGeoHash")


def test_ST_PointFromText():
    _test_geometry_returning_func("ST_PointFromText")


def test_ST_PointFromWKB():
    _test_geometry_returning_func("ST_PointFromWKB")


def test_ST_Polygon():
    _test_geometry_returning_func("ST_Polygon")


def test_ST_PolygonFromText():
    _test_geometry_returning_func("ST_PolygonFromText")


def test_ST_TileEnvelope():
    _test_geometry_returning_func("ST_TileEnvelope")


def test_ST_WKBToSQL():
    _test_geometry_returning_func("ST_WKBToSQL")


def test_ST_WKTToSQL():
    _test_geometry_returning_func("ST_WKTToSQL")


#
# Geometry Accessors
#
def test_ST_Boundary():
    _test_geometry_returning_func("ST_Boundary")


def test_ST_BoundingDiagonal():
    _test_geometry_returning_func("ST_BoundingDiagonal")


def test_ST_EndPoint():
    _test_geometry_returning_func("ST_EndPoint")


def test_ST_Envelope():
    _test_geometry_returning_func("ST_Envelope")


def test_ST_GeometryN():
    _test_geometry_returning_func("ST_GeometryN")


def test_ST_GeometryType():
    _test_simple_func("ST_GeometryType")


def test_ST_InteriorRingN():
    _test_geometry_returning_func("ST_InteriorRingN")


def test_ST_IsValid():
    _test_simple_func("ST_IsValid")


def test_ST_NPoints():
    _test_simple_func("ST_NPoints")


def test_ST_PatchN():
    _test_geometry_returning_func("ST_PatchN")


def test_ST_PointN():
    _test_geometry_returning_func("ST_PointN")


def test_ST_Points():
    _test_geometry_returning_func("ST_Points")


def test_ST_SRID():
    _test_simple_func("ST_SRID")


def test_ST_StartPoint():
    _test_geometry_returning_func("ST_StartPoint")


def test_ST_X():
    _test_simple_func("ST_X")


def test_ST_Y():
    _test_simple_func("ST_Y")


def test_ST_Z():
    _test_simple_func("ST_Z")


#
# Geometry Editors
#
def test_ST_AddPoint():
    _test_geometry_returning_func("ST_AddPoint")


def test_ST_Affine():
    _test_geometry_returning_func("ST_Affine")


def test_ST_CollectionExtract():
    _test_geometry_returning_func("ST_CollectionExtract")


def test_ST_CollectionHomogenize():
    _test_geometry_returning_func("ST_CollectionHomogenize")


def test_ST_ExteriorRing():
    _test_geometry_returning_func("ST_ExteriorRing")


def test_ST_Force2D():
    _test_geometry_returning_func("ST_Force2D")


def test_ST_Force3D():
    _test_geometry_returning_func("ST_Force3D")


def test_ST_Force3DM():
    _test_geometry_returning_func("ST_Force3DM")


def test_ST_Force3DZ():
    _test_geometry_returning_func("ST_Force3DZ")


def test_ST_Force4D():
    _test_geometry_returning_func("ST_Force4D")


def test_ST_ForceCollection():
    _test_geometry_returning_func("ST_ForceCollection")


def test_ST_ForceCurve():
    _test_geometry_returning_func("ST_ForceCurve")


def test_ST_ForcePolygonCCW():
    _test_geometry_returning_func("ST_ForcePolygonCCW")


def test_ST_ForcePolygonCW():
    _test_geometry_returning_func("ST_ForcePolygonCW")


def test_ST_ForceRHR():
    _test_geometry_returning_func("ST_ForceRHR")


def test_ST_ForceSFS():
    _test_geometry_returning_func("ST_ForceSFS")


def test_ST_M():
    _test_simple_func("ST_M")


def test_ST_Multi():
    _test_geometry_returning_func("ST_Multi")


def test_ST_Normalize():
    _test_geometry_returning_func("ST_Normalize")


def test_ST_QuantizeCoordinates():
    _test_geometry_returning_func("ST_QuantizeCoordinates")


def test_ST_RemovePoint():
    _test_geometry_returning_func("ST_RemovePoint")


def test_ST_Reverse():
    _test_geometry_returning_func("ST_Reverse")


def test_ST_Rotate():
    _test_geometry_returning_func("ST_Rotate")


def test_ST_RotateX():
    _test_geometry_returning_func("ST_RotateX")


def test_ST_RotateY():
    _test_geometry_returning_func("ST_RotateY")


def test_ST_RotateZ():
    _test_geometry_returning_func("ST_RotateZ")


def test_ST_Scale():
    _test_geometry_returning_func("ST_Scale")


def test_ST_Segmentize():
    _test_geometry_returning_func("ST_Segmentize")


def test_ST_SetPoint():
    _test_geometry_returning_func("ST_SetPoint")


def test_ST_SetSRID():
    _test_geometry_returning_func("ST_SetSRID")


def test_ST_Snap():
    _test_geometry_returning_func("ST_Snap")


def test_ST_SnapToGrid():
    _test_geometry_returning_func("ST_SnapToGrid")


def test_ST_SwapOrdinates():
    _test_geometry_returning_func("ST_SwapOrdinates")


def test_ST_Transform():
    _test_geometry_returning_func("ST_Transform")


def test_ST_Translate():
    _test_geometry_returning_func("ST_Translate")


def test_ST_TransScale():
    _test_geometry_returning_func("ST_TransScale")


#
# Geometry Outputs
#
def test_ST_AsBinary():
    _test_simple_func("ST_AsBinary")


def test_ST_AsEWKB():
    _test_simple_func("ST_AsEWKB")


def test_ST_AsTWKB():
    _test_simple_func("ST_AsTWKB")


def test_ST_AsGeoJSON():
    _test_simple_func("ST_AsGeoJSON")


def test_ST_AsGML():
    _test_simple_func("ST_AsGML")


def test_ST_AsKML():
    _test_simple_func("ST_AsKML")


def test_ST_AsSVG():
    _test_simple_func("ST_AsSVG")


def test_ST_AsText():
    _test_simple_func("ST_AsText")


def test_ST_AsEWKT():
    _test_simple_func("ST_AsEWKT")


def test_ST_AsMVTGeom():
    _test_geometry_returning_func("ST_AsMVTGeom")


#
# Spatial Relationships and Measurements
#
def test_ST_Area():
    _test_simple_func("ST_Area")


def test_ST_Azimuth():
    _test_simple_func("ST_Azimuth")


def test_ST_Centroid():
    _test_geometry_returning_func("ST_Centroid")


def test_ST_ClosestPoint():
    _test_geometry_returning_func("ST_ClosestPoint")


def test_ST_3DClosestPoint():
    _test_geometry_returning_func("ST_3DClosestPoint")


def test_ST_Contains():
    _test_simple_func("ST_Contains")


def test_ST_ContainsProperly():
    _test_simple_func("ST_ContainsProperly")


def test_ST_Covers():
    _test_simple_func("ST_Covers")


def test_ST_CoveredBy():
    _test_simple_func("ST_CoveredBy")


def test_ST_Crosses():
    _test_simple_func("ST_Crosses")


def test_ST_Disjoint():
    _test_simple_func("ST_Disjoint")


def test_ST_Distance():
    _test_simple_func("ST_Distance")


def test_ST_Distance_Sphere():
    _test_simple_func("ST_Distance_Sphere")


def test_ST_DistanceSphere():
    _test_simple_func("ST_DistanceSphere")


def test_ST_DFullyWithin():
    _test_simple_func("ST_DFullyWithin")


def test_ST_DWithin():
    _test_simple_func("ST_DWithin")


def test_ST_Equals():
    _test_simple_func("ST_Equals")


def test_ST_Intersects():
    _test_simple_func("ST_Intersects")


def test_ST_Length():
    _test_simple_func("ST_Length")


def test_ST_LineLocatePoint():
    _test_simple_func("ST_LineLocatePoint")


def test_ST_LongestLine():
    _test_geometry_returning_func("ST_LongestLine")


def test_ST_3DLongestLine():
    _test_geometry_returning_func("ST_3DLongestLine")


def test_ST_MinimumClearanceLine():
    _test_geometry_returning_func("ST_MinimumClearanceLine")


def test_ST_OrderingEquals():
    _test_simple_func("ST_OrderingEquals")


def test_ST_Overlaps():
    _test_simple_func("ST_Overlaps")


def test_ST_Perimeter():
    _test_simple_func("ST_Perimeter")


def test_ST_Project():
    _test_geography_returning_func("ST_Project")


def test_ST_Relate():
    _test_simple_func("ST_Relate")


def test_ST_ShortestLine():
    _test_geometry_returning_func("ST_ShortestLine")


def test_ST_3DShortestLine():
    _test_geometry_returning_func("ST_3DShortestLine")


def test_ST_Touches():
    _test_simple_func("ST_Touches")


def test_ST_Within():
    _test_simple_func("ST_Within")


#
# Geometry Processing
#
def test_ST_Buffer():
    _test_geometry_returning_func("ST_Buffer")


def test_ST_BuildArea():
    _test_geometry_returning_func("ST_BuildArea")


def test_ST_ClipByBox2D():
    _test_geometry_returning_func("ST_ClipByBox2D")


def test_ST_ChaikinSmoothing():
    _test_geometry_returning_func("ST_ChaikinSmoothing")


def test_ST_ConcaveHull():
    _test_geometry_returning_func("ST_ConcaveHull")


def test_ST_ConvexHull():
    _test_geometry_returning_func("ST_ConvexHull")


def test_ST_CurveToLine():
    _test_geometry_returning_func("ST_CurveToLine")


def test_ST_DelaunayTriangles():
    _test_geometry_returning_func("ST_DelaunayTriangles")


def test_ST_Difference():
    _test_geometry_returning_func("ST_Difference")


def test_ST_Dump():
    _test_simple_func("ST_Dump")


def test_ST_DumpPoints():
    _test_simple_func("ST_DumpPoints")


def test_ST_FilterByM():
    _test_geometry_returning_func("ST_FilterByM")


def test_ST_FlipCoordinates():
    _test_geometry_returning_func("ST_FlipCoordinates")


def test_ST_GeneratePoints():
    _test_geometry_returning_func("ST_GeneratePoints")


def test_ST_GeometricMedian():
    _test_geometry_returning_func("ST_GeometricMedian")


def test_ST_Intersection():
    _test_geometry_returning_func("ST_Intersection")


def test_ST_LineToCurve():
    _test_geometry_returning_func("ST_LineToCurve")


def test_ST_LineMerge():
    _test_geometry_returning_func("ST_LineMerge")


def test_ST_LineSubstring():
    _test_geometry_returning_func("ST_LineSubstring")


def test_ST_MakeValid():
    _test_geometry_returning_func("ST_MakeValid")


def test_ST_MemUnion():
    _test_geometry_returning_func("ST_MemUnion")


def test_ST_MinimumBoundingCircle():
    _test_geometry_returning_func("ST_MinimumBoundingCircle")


def test_ST_Node():
    _test_geometry_returning_func("ST_Node")


def test_ST_OffsetCurve():
    _test_geometry_returning_func("ST_OffsetCurve")


def test_ST_OrientedEnvelope():
    _test_geometry_returning_func("ST_OrientedEnvelope")


def test_ST_PointOnSurface():
    _test_geometry_returning_func("ST_PointOnSurface")


def test_ST_Polygonize():
    _test_geometry_returning_func("ST_Polygonize")


def test_ST_RemoveRepeatedPoints():
    _test_geometry_returning_func("ST_RemoveRepeatedPoints")


def test_ST_SetEffectiveArea():
    _test_geometry_returning_func("ST_SetEffectiveArea")


def test_ST_SharedPaths():
    _test_geometry_returning_func("ST_SharedPaths")


def test_ST_ShiftLongitude():
    _test_geometry_returning_func("ST_ShiftLongitude")


def test_ST_Simplify():
    _test_geometry_returning_func("ST_Simplify")


def test_ST_SimplifyPreserveTopology():
    _test_geometry_returning_func("ST_SimplifyPreserveTopology")


def test_ST_SimplifyVW():
    _test_geometry_returning_func("ST_SimplifyVW")


def test_ST_Split():
    _test_geometry_returning_func("ST_Split")


def test_ST_Subdivide():
    _test_geometry_returning_func("ST_Subdivide")


def test_ST_SymDifference():
    _test_geometry_returning_func("ST_SymDifference")


def test_ST_Union():
    _test_geometry_returning_func("ST_Union")


def test_ST_UnaryUnion():
    _test_geometry_returning_func("ST_UnaryUnion")


def test_ST_VoronoiLines():
    _test_geometry_returning_func("ST_VoronoiLines")


def test_ST_VoronoiPolygons():
    _test_geometry_returning_func("ST_VoronoiPolygons")


def test_ST_WrapX():
    _test_geometry_returning_func("ST_WrapX")


#
# Bounding Box Functions
#
def test_ST_Expand():
    _test_geometry_returning_func("ST_Expand")


#
# Linear Referencing
#
def test_ST_AddMeasure():
    _test_geometry_returning_func("ST_AddMeasure")


def test_ST_LineInterpolatePoint():
    _test_geometry_returning_func("ST_LineInterpolatePoint")


def test_ST_LineInterpolatePoints():
    _test_geometry_returning_func("ST_LineInterpolatePoints")


def test_ST_LocateAlong():
    _test_geometry_returning_func("ST_LocateAlong")


def test_ST_LocateBetween():
    _test_geometry_returning_func("ST_LocateBetween")


def test_ST_LocateBetweenElevations():
    _test_geometry_returning_func("ST_LocateBetweenElevations")


def test_ST_3DLineInterpolatePoint():
    _test_geometry_returning_func("ST_3DLineInterpolatePoint")


#
# Raster Constructors
#
def test_ST_AddBand():
    _test_raster_returning_func("ST_AddBand")


def test_ST_AsRaster():
    _test_raster_returning_func("ST_AsRaster")


#
# Raster Editors
#
def test_ST_Resample():
    _test_raster_returning_func("ST_Resample")


def test_ST_Rescale():
    _test_raster_returning_func("ST_Rescale")


def test_ST_Reskew():
    _test_raster_returning_func("ST_Reskew")


# ST_SnapToGrid already exists for Geometry type so it can not be duplicated
def test_ST_SnapToGrid_raster():
    _test_raster_returning_func("ST_SnapToGrid", type_=Raster)


def test_ST_Resize():
    _test_raster_returning_func("ST_Resize")


#
# Raster Accessors
#
def test_ST_Height():
    _test_simple_func("ST_Height")


def test_ST_Width():
    _test_simple_func("ST_Width")


#
# Raster Pixel Accessors and Setters
#
def test_ST_Value():
    _test_simple_func("ST_Value")


#
# Raster Band Statistics and Analytics
#
def test_ST_ValueCount():
    _test_simple_func("ST_ValueCount")


#
# DEM (Elevation)
#
def test_ST_HillShade():
    _test_raster_returning_func("ST_HillShade")



"""This module defines the Column types.

The :class:`geoalchemy2.types.Geometry`, :class:`geoalchemy2.types.Geography`, and
:class:`geoalchemy2.types.Raster` classes are used when defining geometry, geography and raster
columns/properties in models.
"""
import warnings
from typing import Any
from typing import Dict
from typing import Optional

from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql.base import ischema_names as _postgresql_ischema_names
from sqlalchemy.dialects.sqlite.base import ischema_names as _sqlite_ischema_names
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import func
from sqlalchemy.types import Float
from sqlalchemy.types import Integer
from sqlalchemy.types import UserDefinedType

try:
    # SQLAlchemy >= 2
    from sqlalchemy.sql._typing import _TypeEngineArgument
except ImportError:
    # SQLAlchemy < 2
    _TypeEngineArgument = Any  # type: ignore

from geoalchemy2.comparator import BaseComparator
from geoalchemy2.comparator import Comparator
from geoalchemy2.elements import CompositeElement
from geoalchemy2.elements import RasterElement
from geoalchemy2.elements import WKBElement
from geoalchemy2.exc import ArgumentError
from geoalchemy2.types import dialects


def select_dialect(dialect_name):
    """Select the dialect from its name."""
    known_dialects = {
        "geopackage": dialects.geopackage,
        "mysql": dialects.mysql,
        "postgresql": dialects.postgresql,
        "sqlite": dialects.sqlite,
    }
    return known_dialects.get(dialect_name, dialects.common)


class _GISType(UserDefinedType):
    """The base class for spatial types.

    This class defines ``bind_expression`` and ``column_expression`` methods
    that wrap column expressions in ``ST_GeomFromEWKT``, ``ST_GeogFromText``,
    or ``ST_AsEWKB`` calls.

    This class also defines ``result_processor`` and ``bind_processor``
    methods. The function returned by ``result_processor`` converts WKB values
    received from the database to :class:`geoalchemy2.elements.WKBElement`
    objects. The function returned by ``bind_processor`` converts
    :class:`geoalchemy2.elements.WKTElement` objects to EWKT strings.

    Args:
        geometry_type: The geometry type.

            Possible values are:

              * ``"GEOMETRY"``,
              * ``"POINT"``,
              * ``"LINESTRING"``,
              * ``"POLYGON"``,
              * ``"MULTIPOINT"``,
              * ``"MULTILINESTRING"``,
              * ``"MULTIPOLYGON"``,
              * ``"GEOMETRYCOLLECTION"``,
              * ``"CURVE"``,
              * ``None``.

           The latter is actually not supported with
           :class:`geoalchemy2.types.Geography`.

           When set to ``None`` then no "geometry type" constraints will be
           attached to the geometry type declaration.

           Default is ``"GEOMETRY"``.

        srid: The SRID for this column. E.g. 4326. Default is ``-1``.
        dimension: The dimension of the geometry. Default is ``2``.
        spatial_index: Indicate if a spatial index should be created. Default is ``True``.
        use_N_D_index: Use the N-D index instead of the standard 2-D index.
        use_typmod: By default PostgreSQL type modifiers are used to create the geometry
            column. To use check constraints instead set ``use_typmod`` to
            ``False``. By default this option is not included in the call to
            ``AddGeometryColumn``. Note that this option is only available for PostGIS 2.x.
    """

    name: Optional[str] = None
    """ Name used for defining the main geo type (geometry or geography)
        in CREATE TABLE statements. Set in subclasses. """

    from_text: Optional[str] = None
    """ The name of "from text" function for this type.
        Set in subclasses. """

    as_binary: Optional[str] = None
    """ The name of the "as binary" function for this type.
        Set in subclasses. """

    comparator_factory: Any = Comparator
    """ This is the way by which spatial operators are defined for
        geometry/geography columns. """

    cache_ok = False
    """ Disable cache for this type. """

    def __init__(
        self,
        geometry_type="GEOMETRY",
        srid=-1,
        dimension=2,
        spatial_index=True,
        use_N_D_index=False,
        use_typmod=None,
        from_text=None,
        name=None,
        nullable=True,
        _spatial_index_reflected=None,
    ) -> None:
        geometry_type, srid = self.check_ctor_args(
            geometry_type, srid, dimension, use_typmod, nullable
        )
        self.geometry_type = geometry_type
        self.srid = srid
        if name is not None:
            self.name = name
        if from_text is not None:
            self.from_text = from_text
        self.dimension = dimension
        self.spatial_index = spatial_index
        self.use_N_D_index = use_N_D_index
        self.use_typmod = use_typmod
        self.extended: Optional[bool] = self.as_binary == "ST_AsEWKB"
        self.nullable = nullable
        self._spatial_index_reflected = _spatial_index_reflected

    def get_col_spec(self):
        if not self.geometry_type:
            return self.name
        return "%s(%s,%d)" % (self.name, self.geometry_type, self.srid)

    def column_expression(self, col):
        """Specific column_expression that automatically adds a conversion function."""
        return getattr(func, self.as_binary)(col, type_=self)

    def result_processor(self, dialect, coltype):
        """Specific result_processor that automatically process spatial elements."""

        def process(value):
            if value is not None:
                kwargs = {}
                if self.srid > 0:
                    kwargs["srid"] = self.srid
                if self.extended is not None and dialect.name != "mysql":
                    kwargs["extended"] = self.extended
                return self.ElementType(value, **kwargs)

        return process

    def bind_expression(self, bindvalue):
        """Specific bind_expression that automatically adds a conversion function."""
        return getattr(func, self.from_text)(bindvalue, type_=self)

    def bind_processor(self, dialect):
        """Specific bind_processor that automatically process spatial elements."""

        def process(bindvalue):
            return select_dialect(dialect.name).bind_processor_process(self, bindvalue)

        return process

    @staticmethod
    def check_ctor_args(geometry_type, srid, dimension, use_typmod, nullable):
        try:
            srid = int(srid)
        except ValueError:
            raise ArgumentError("srid must be convertible to an integer")
        if geometry_type:
            geometry_type = geometry_type.upper()
        elif srid > 0:
            warnings.warn("srid not enforced when geometry_type is None")

        if use_typmod is not None and not nullable:
            raise ArgumentError(
                'The "nullable" and "use_typmod" arguments can not be used together'
            )

        return geometry_type, srid


@compiles(_GISType, "mysql")
def get_col_spec(self, *args, **kwargs):
    if self.geometry_type is not None:
        spec = "%s" % self.geometry_type
    else:
        spec = "GEOMETRY"

    if not self.nullable or self.spatial_index:
        spec += " NOT NULL"
    if self.srid > 0:
        spec += " SRID %d" % self.srid
    return spec


class Geometry(_GISType):
    """The Geometry type.

    Creating a geometry column is done like this::

        Column(Geometry(geometry_type='POINT', srid=4326))

    See :class:`geoalchemy2.types._GISType` for the list of arguments that can
    be passed to the constructor.

    If ``srid`` is set then the ``WKBElement`` objects resulting from queries will
    have that SRID, and, when constructing the ``WKBElement`` objects, the SRID
    won't be read from the data returned by the database. If ``srid`` is not set
    (meaning it's ``-1``) then the SRID set in ``WKBElement`` objects will be read
    from the data returned by the database.
    """

    name = "geometry"
    """ Type name used for defining geometry columns in ``CREATE TABLE``. """

    from_text = "ST_GeomFromEWKT"
    """ The "from text" geometry constructor. Used by the parent class'
        ``bind_expression`` method. """

    as_binary = "ST_AsEWKB"
    """ The "as binary" function to use. Used by the parent class'
        ``column_expression`` method. """

    ElementType = WKBElement
    """ The element class to use. Used by the parent class'
        ``result_processor`` method. """

    cache_ok = False
    """ Disable cache for this type. """


class Geography(_GISType):
    """The Geography type.

    Creating a geography column is done like this::

        Column(Geography(geometry_type='POINT', srid=4326))

    See :class:`geoalchemy2.types._GISType` for the list of arguments that can
    be passed to the constructor.
    """

    name = "geography"
    """ Type name used for defining geography columns in ``CREATE TABLE``. """

    from_text = "ST_GeogFromText"
    """ The ``FromText`` geography constructor. Used by the parent class'
        ``bind_expression`` method. """

    as_binary = "ST_AsBinary"
    """ The "as binary" function to use. Used by the parent class'
        ``column_expression`` method. """

    ElementType = WKBElement
    """ The element class to use. Used by the parent class'
        ``result_processor`` method. """

    cache_ok = False
    """ Disable cache for this type. """


class Raster(_GISType):
    """The Raster column type.

    Creating a raster column is done like this::

        Column(Raster)

    This class defines the ``result_processor`` method, so that raster values
    received from the database are converted to
    :class:`geoalchemy2.elements.RasterElement` objects.

    Args:
        spatial_index: Indicate if a spatial index should be created. Default is ``True``.
    """

    comparator_factory = BaseComparator
    """
    This is the way by which spatial operators and functions are
    defined for raster columns.
    """

    name = "raster"
    """ Type name used for defining raster columns in ``CREATE TABLE``. """

    from_text = "raster"
    """ The "from text" raster constructor. Used by the parent class'
        ``bind_expression`` method. """

    as_binary = "raster"
    """ The "as binary" function to use. Used by the parent class'
        ``column_expression`` method. """

    ElementType = RasterElement
    """ The element class to use. Used by the parent class'
        ``result_processor`` method. """

    cache_ok = False
    """ Disable cache for this type. """

    def __init__(self, spatial_index=True, from_text=None, name=None, nullable=True) -> None:
        # Enforce default values
        super(Raster, self).__init__(
            geometry_type=None,
            srid=-1,
            dimension=2,
            spatial_index=spatial_index,
            use_N_D_index=False,
            use_typmod=False,
            from_text=from_text,
            name=name,
            nullable=nullable,
        )
        self.extended = None

    @staticmethod
    def check_ctor_args(*args, **kwargs):
        return None, -1


class _DummyGeometry(Geometry):
    """A dummy type only used with SQLite."""

    def get_col_spec(self):
        return self.geometry_type or "GEOMETRY"


class CompositeType(UserDefinedType):
    """A composite type used by some spatial functions.

    A wrapper for :class:`geoalchemy2.elements.CompositeElement`, that can be
    used as the return type in PostgreSQL functions that return composite
    values.

    This is used as the base class of :class:`geoalchemy2.types.GeometryDump`.
    """

    typemap: Dict[str, _TypeEngineArgument] = {}
    """ Dictionary used for defining the content types and their
        corresponding keys. Set in subclasses. """

    class comparator_factory(UserDefinedType.Comparator):
        def __getattr__(self, key):
            try:
                type_ = self.type.typemap[key]
            except KeyError:
                raise AttributeError("Type '%s' doesn't have an attribute: '%s'" % (self.type, key))

            return CompositeElement(self.expr, key, type_)


class GeometryDump(CompositeType):
    """The return type for functions like ``ST_Dump``.

    The type consists in a path and a geom field.
    You should normally never use this class directly.
    """

    typemap = {"path": postgresql.ARRAY(Integer), "geom": Geometry}
    """ Dictionary defining the contents of a ``geometry_dump``. """

    cache_ok = True
    """ Enable cache for this type. """


# Register Geometry, Geography and Raster to SQLAlchemy's reflection subsystems.
_postgresql_ischema_names["geometry"] = Geometry
_postgresql_ischema_names["geography"] = Geography
_postgresql_ischema_names["raster"] = Raster

_sqlite_ischema_names["GEOMETRY"] = Geometry
_sqlite_ischema_names["POINT"] = Geometry
_sqlite_ischema_names["LINESTRING"] = Geometry
_sqlite_ischema_names["POLYGON"] = Geometry
_sqlite_ischema_names["MULTIPOINT"] = Geometry
_sqlite_ischema_names["MULTILINESTRING"] = Geometry
_sqlite_ischema_names["MULTIPOLYGON"] = Geometry
_sqlite_ischema_names["CURVE"] = Geometry
_sqlite_ischema_names["GEOMETRYCOLLECTION"] = Geometry
_sqlite_ischema_names["RASTER"] = Raster


class SummaryStats(CompositeType):
    """Define the composite type returned by the function ST_SummaryStatsAgg."""

    typemap = {
        "count": Integer,
        "sum": Float,
        "mean": Float,
        "stddev": Float,
        "min": Float,
        "max": Float,
    }

    cache_ok = True
    """ Enable cache for this type. """


__all__ = [
    "_GISType",
    "CompositeType",
    "Geography",
    "Geometry",
    "GeometryDump",
    "Raster",
    "SummaryStats",
    "dialects",
    "select_dialect",
]


def __dir__():
    return __all__



import re

import pytest
from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import Table

from geoalchemy2.elements import WKBElement
from geoalchemy2.elements import WKTElement
from geoalchemy2.types import Geometry

from . import select


def eq_sql(a, b):
    a = re.sub(r"[\n\t]", "", str(a))
    assert a == b


@pytest.fixture
def geometry_table():
    table = Table("table", MetaData(), Column("geom", Geometry))
    return table


@pytest.fixture(params=["raw_str", "WKTElement", "WKBElement"])
def point_reference(request):
    wkt = "POINT(1 2)"
    if request.param == "WKTElement":
        return WKTElement(wkt)
    elif request.param == "WKBElement":
        return WKBElement(
            b"\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\xf0?\x00\x00\x00\x00\x00\x00\x00@"
        )
    elif request.param == "raw_str":
        return wkt


class TestOperator:
    def test_eq(self, geometry_table, point_reference):
        expr = geometry_table.c.geom == point_reference
        eq_sql(expr, '"table".geom = ST_GeomFromEWKT(:geom_1)')

    def test_eq_with_None(self, geometry_table):
        expr = geometry_table.c.geom == None  # NOQA
        eq_sql(expr, '"table".geom IS NULL')

    def test_ne(self, geometry_table, point_reference):
        expr = geometry_table.c.geom != point_reference
        eq_sql(expr, '"table".geom != ST_GeomFromEWKT(:geom_1)')

    def test_ne_with_None(self, geometry_table):
        expr = geometry_table.c.geom != None  # NOQA
        eq_sql(expr, '"table".geom IS NOT NULL')

    def test_intersects(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.intersects(point_reference)
        eq_sql(expr, '"table".geom && ST_GeomFromEWKT(:geom_1)')

    def test_overlaps_or_to_left(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.overlaps_or_to_left(point_reference)
        eq_sql(expr, '"table".geom &< ST_GeomFromEWKT(:geom_1)')

    def test_overlaps_or_below(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.overlaps_or_below(point_reference)
        eq_sql(expr, '"table".geom &<| ST_GeomFromEWKT(:geom_1)')

    def test_overlaps_or_to_right(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.overlaps_or_to_right(point_reference)
        eq_sql(expr, '"table".geom &> ST_GeomFromEWKT(:geom_1)')

    def test_to_left(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.to_left(point_reference)
        eq_sql(expr, '"table".geom << ST_GeomFromEWKT(:geom_1)')

    def test_lshift(self, geometry_table, point_reference):
        expr = geometry_table.c.geom << point_reference
        eq_sql(expr, '"table".geom << ST_GeomFromEWKT(:geom_1)')

    def test_below(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.below(point_reference)
        eq_sql(expr, '"table".geom <<| ST_GeomFromEWKT(:geom_1)')

    def test_to_right(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.to_right(point_reference)
        eq_sql(expr, '"table".geom >> ST_GeomFromEWKT(:geom_1)')

    def test_rshift(self, geometry_table, point_reference):
        expr = geometry_table.c.geom >> point_reference
        eq_sql(expr, '"table".geom >> ST_GeomFromEWKT(:geom_1)')

    def test_contained(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.contained(point_reference)
        eq_sql(expr, '"table".geom @ ST_GeomFromEWKT(:geom_1)')

    def test_overlaps_or_above(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.overlaps_or_above(point_reference)
        eq_sql(expr, '"table".geom |&> ST_GeomFromEWKT(:geom_1)')

    def test_above(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.above(point_reference)
        eq_sql(expr, '"table".geom |>> ST_GeomFromEWKT(:geom_1)')

    def test_contains(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.contains(point_reference)
        eq_sql(expr, '"table".geom ~ ST_GeomFromEWKT(:geom_1)')

    def test_same(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.same(point_reference)
        eq_sql(expr, '"table".geom ~= ST_GeomFromEWKT(:geom_1)')

    def test_distance_centroid(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.distance_centroid(point_reference)
        eq_sql(expr, '"table".geom <-> ST_GeomFromEWKT(:geom_1)')

    def test_distance_centroid_select(self, geometry_table, point_reference):
        s = (
            geometry_table.select()
            .order_by(geometry_table.c.geom.distance_centroid(point_reference))
            .limit(10)
        )
        eq_sql(
            s,
            'SELECT ST_AsEWKB("table".geom) AS geom '
            'FROM "table" '
            'ORDER BY "table".geom <-> ST_GeomFromEWKT(:geom_1) '
            "LIMIT :param_1",
        )
        assert s.compile().params == {"geom_1": point_reference, "param_1": 10}

    def test_distance_centroid_select_with_label(self, geometry_table, point_reference):
        s = select([geometry_table.c.geom.distance_centroid(point_reference).label("dc")])
        s = s.order_by("dc").limit(10)
        eq_sql(
            s,
            'SELECT "table".geom <-> ST_GeomFromEWKT(:geom_1) AS dc '
            'FROM "table" ORDER BY dc LIMIT :param_1',
        )
        assert s.compile().params == {"geom_1": point_reference, "param_1": 10}

    def test_distance_box(self, geometry_table, point_reference):
        expr = geometry_table.c.geom.distance_box(point_reference)
        eq_sql(expr, '"table".geom <#> ST_GeomFromEWKT(:geom_1)')

    def test_distance_box_select(self, geometry_table, point_reference):
        s = (
            geometry_table.select()
            .order_by(geometry_table.c.geom.distance_box(point_reference))
            .limit(10)
        )
        eq_sql(
            s,
            'SELECT ST_AsEWKB("table".geom) AS geom '
            'FROM "table" '
            'ORDER BY "table".geom <#> ST_GeomFromEWKT(:geom_1) '
            "LIMIT :param_1",
        )
        assert s.compile().params == {"geom_1": point_reference, "param_1": 10}

    def test_distance_box_select_with_label(self, geometry_table, point_reference):
        s = select([geometry_table.c.geom.distance_box(point_reference).label("dc")])
        s = s.order_by("dc").limit(10)
        eq_sql(
            s,
            'SELECT "table".geom <#> ST_GeomFromEWKT(:geom_1) AS dc '
            'FROM "table" ORDER BY dc LIMIT :param_1',
        )
        assert s.compile().params == {"geom_1": point_reference, "param_1": 10}

    def test_intersects_nd(self, geometry_table):
        expr = geometry_table.c.geom.intersects_nd(
            "Box3D(ST_GeomFromEWKT('LINESTRING(1 2 3, 3 4 5, 5 6 5)'));"
        )
        eq_sql(expr, '"table".geom &&& ST_GeomFromEWKT(:geom_1)')



"""This module defines the functions used for administration tasks."""
from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy import Table
from sqlalchemy import event
from sqlalchemy.sql import func

from geoalchemy2.admin import dialects
from geoalchemy2.admin.dialects.common import _check_spatial_type
from geoalchemy2.admin.dialects.common import _spatial_idx_name
from geoalchemy2.exc import ArgumentError
from geoalchemy2.types import Geography
from geoalchemy2.types import Geometry
from geoalchemy2.types import Raster


def select_dialect(dialect_name):
    """Select the dialect from its name."""
    known_dialects = {
        "geopackage": dialects.geopackage,
        "mysql": dialects.mysql,
        "postgresql": dialects.postgresql,
        "sqlite": dialects.sqlite,
    }
    return known_dialects.get(dialect_name, dialects.common)


def setup_ddl_event_listeners():
    """Setup the DDL event listeners to automatically process spatial columns."""

    @event.listens_for(Table, "before_create")
    def before_create(table, bind, **kw):
        """Handle spatial indexes."""
        select_dialect(bind.dialect.name).before_create(table, bind, **kw)

    @event.listens_for(Table, "after_create")
    def after_create(table, bind, **kw):
        """Restore original column list including managed Geometry columns."""
        select_dialect(bind.dialect.name).after_create(table, bind, **kw)

    @event.listens_for(Table, "before_drop")
    def before_drop(table, bind, **kw):
        """Drop the managed Geometry columns."""
        select_dialect(bind.dialect.name).before_drop(table, bind, **kw)

    @event.listens_for(Table, "after_drop")
    def after_drop(table, bind, **kw):
        """Restore original column list including managed Geometry columns."""
        select_dialect(bind.dialect.name).after_drop(table, bind, **kw)

    @event.listens_for(Column, "after_parent_attach")
    def after_parent_attach(column, table):
        """Automatically add spatial indexes."""
        if not isinstance(table, Table):
            # For old versions of SQLAlchemy, subqueries might trigger the after_parent_attach event
            # with a selectable as table, so we want to skip this case.
            return

        if not getattr(column.type, "nullable", True):
            column.nullable = column.type.nullable
        elif hasattr(column.type, "nullable"):
            column.type.nullable = column.nullable

        if not getattr(column.type, "spatial_index", False) and getattr(
            column.type, "use_N_D_index", False
        ):
            raise ArgumentError("Arg Error(use_N_D_index): spatial_index must be True")

        if not getattr(column.type, "spatial_index", False):
            # If the column is managed, the indexes are created after the table
            return

        try:
            if column.type._spatial_index_reflected:
                return
        except AttributeError:
            pass

        kwargs = {
            "postgresql_using": "gist",
            "_column_flag": True,
        }
        col = column
        if _check_spatial_type(column.type, (Geometry, Geography)):
            if column.type.use_N_D_index:
                kwargs["postgresql_ops"] = {column.name: "gist_geometry_ops_nd"}
        elif _check_spatial_type(column.type, Raster):
            col = func.ST_ConvexHull(column)

        table.append_constraint(
            Index(
                _spatial_idx_name(table.name, column.name),
                col,
                **kwargs,
            )
        )

    @event.listens_for(Table, "column_reflect")
    def _reflect_geometry_column(inspector, table, column_info):
        select_dialect(inspector.bind.dialect.name).reflect_geometry_column(
            inspector, table, column_info
        )


__all__ = [
    "dialects",
    "select_dialect",
    "setup_ddl_event_listeners",
]


def __dir__():
    return __all__



"""This module defines specific functions for GeoPackage dialect.

See GeoPackage specifications here: http://www.geopackage.org/spec/
"""
import re

from sqlalchemy import text
from sqlalchemy.dialects import registry
from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import func
from sqlalchemy.sql import select

from geoalchemy2 import functions
from geoalchemy2.admin.dialects.common import _check_spatial_type
from geoalchemy2.admin.dialects.common import _format_select_args
from geoalchemy2.admin.dialects.common import _spatial_idx_name
from geoalchemy2.admin.dialects.common import setup_create_drop
from geoalchemy2.admin.dialects.sqlite import _SQLITE_FUNCTIONS
from geoalchemy2.admin.dialects.sqlite import get_col_dim
from geoalchemy2.admin.dialects.sqlite import load_spatialite_driver
from geoalchemy2.types import Geography
from geoalchemy2.types import Geometry
from geoalchemy2.types import _DummyGeometry


class GeoPackageDialect(SQLiteDialect_pysqlite):
    """Define a specific dialect for GeoPackage."""

    name = "geopackage"
    driver = "gpkg"

    supports_statement_cache = True
    """Enable caching for GeoPackage dialect."""


registry.register("gpkg", "geoalchemy2.admin.dialects.geopackage", "GeoPackageDialect")


def load_geopackage_driver(dbapi_conn, *args):
    """Load SpatiaLite extension in GeoPackage connection and set VirtualGpkg and Amphibious modes.

    .. Warning::
        The path to the SpatiaLite module should be set in the `SPATIALITE_LIBRARY_PATH`
        environment variable.

    Args:
        dbapi_conn: The DBAPI connection.
    """
    load_spatialite_driver(dbapi_conn, *args)

    dbapi_conn.execute("SELECT AutoGpkgStart();")
    dbapi_conn.execute("SELECT EnableGpkgAmphibiousMode();")


def init_geopackage(dbapi_conn, *args):
    """Initialize GeoPackage tables.

    Args:
        dbapi_conn: The DBAPI connection.

    .. Warning::
        No EPSG SRID is loaded in the `gpkg_spatial_ref_sys` table after initialization but
        it is possible to load other EPSG SRIDs afterwards using the
        `gpkgInsertEpsgSRID(srid)`.
        Nevertheless, SRIDs of newly created tables are automatically added.
    """
    if not dbapi_conn.execute("SELECT CheckGeoPackageMetaData();").fetchone()[0]:
        # This only works on the main database
        dbapi_conn.execute("SELECT gpkgCreateBaseTables();")


def load_spatialite_gpkg(*args, **kwargs):
    """Load SpatiaLite extension in GeoPackage and initialize internal tables.

    See :func:`geoalchemy2.admin.dialects.geopackage.load_geopackage_driver` and
    :func:`geoalchemy2.admin.dialects.geopackage.init_geopackage` functions for details about
    arguments.
    """
    load_geopackage_driver(*args)
    init_geopackage(*args, **kwargs)


def _get_spatialite_attrs(bind, table_name, col_name):
    attrs = bind.execute(
        text(
            """SELECT
                A.geometry_type_name,
                A.srs_id,
                A.z,
                A.m,
                IFNULL(B.has_index, 0) AS has_index
            FROM gpkg_geometry_columns
            AS A
            LEFT JOIN (
                SELECT table_name, column_name, COUNT(*) AS has_index
                FROM gpkg_extensions
                WHERE LOWER(table_name) = LOWER(:table_name)
                    AND column_name = :column_name
                    AND extension_name = 'gpkg_rtree_index'
            ) AS B
            ON LOWER(A.table_name) = LOWER(B.table_name)
                AND A.column_name = B.column_name
            WHERE LOWER(A.table_name) = LOWER(:table_name)
                AND A.column_name = :column_name;
        """
        ).bindparams(table_name=table_name, column_name=col_name)
    ).fetchone()
    if attrs is None:
        # If the column is not registered as a spatial column we ignore it
        return None
    geometry_type, srid, has_z, has_m, has_index = attrs
    coord_dimension = "XY"
    if has_z:
        coord_dimension += "Z"
    if has_m:
        coord_dimension += "M"
    col_attributes = geometry_type, coord_dimension, srid, has_index
    return col_attributes


def _setup_dummy_type(table, gis_cols):
    """Setup dummy type for new Geometry columns so they can be updated later."""
    for col in gis_cols:
        # Add dummy columns with GEOMETRY type
        type_str = re.fullmatch("(.+?)[ZMzm]*", col.type.geometry_type).group(1)
        col._actual_type = col.type
        col.type = _DummyGeometry(geometry_type=type_str)
    table.columns = table.info["_saved_columns"]


def create_spatial_index(bind, table, col):
    """Create spatial index on the given column."""
    stmt = select(*_format_select_args(func.gpkgAddSpatialIndex(table.name, col.name)))
    stmt = stmt.execution_options(autocommit=True)
    bind.execute(stmt)


def disable_spatial_index(bind, table, col):
    """Disable spatial indexes if present."""
    for i in ["", "_node", "_parent", "_rowid"]:
        bind.execute(
            text(
                "DROP TABLE IF EXISTS rtree_{}_{}{};".format(
                    table.name,
                    col.name,
                    i,
                )
            )
        )
    bind.execute(
        text(
            """DELETE FROM gpkg_extensions
            WHERE LOWER(table_name) = LOWER(:table_name)
                AND column_name = :column_name
                AND extension_name = 'gpkg_rtree_index';"""
        ).bindparams(
            table_name=table.name,
            column_name=col.name,
        )
    )


def reflect_geometry_column(inspector, table, column_info):
    """Reflect a column of type Geometry with GeoPackage dialect."""
    # Get geometry type, SRID and spatial index from the SpatiaLite metadata
    if not isinstance(column_info.get("type"), Geometry):
        return
    col_attributes = _get_spatialite_attrs(inspector.bind, table.name, column_info["name"])
    if col_attributes is not None:
        geometry_type, coord_dimension, srid, spatial_index = col_attributes

        coord_dimension = {
            "XY": 2,
            "XYZ": 3,
            "XYM": 3,
            "XYZM": 4,
        }.get(coord_dimension, coord_dimension)

        # Set attributes
        column_info["type"].geometry_type = geometry_type
        column_info["type"].dimension = coord_dimension
        column_info["type"].srid = srid
        column_info["type"].spatial_index = bool(spatial_index)

        # Spatial indexes are not automatically reflected with GeoPackage dialect
        column_info["type"]._spatial_index_reflected = False


def before_create(table, bind, **kw):
    """Handle spatial indexes during the before_create event."""
    dialect, gis_cols, regular_cols = setup_create_drop(table, bind)

    # Remove the spatial indexes from the table metadata because they should not be
    # created during the table.create() step since the associated columns do not exist
    # at this time.
    table.info["_after_create_indexes"] = []
    current_indexes = set(table.indexes)
    for idx in current_indexes:
        for col in table.info["_saved_columns"]:
            if _check_spatial_type(col.type, Geometry, dialect) and col in idx.columns.values():
                table.indexes.remove(idx)
                if idx.name != _spatial_idx_name(table.name, col.name) or not getattr(
                    col.type, "spatial_index", False
                ):
                    table.info["_after_create_indexes"].append(idx)

    if len(gis_cols) > 1:
        raise ValueError("Only one geometry column is allowed for a table stored in a GeoPackage.")
    elif len(gis_cols) == 1:
        col = gis_cols[0]
        srid = col.type.srid

        if col.type.geometry_type is None:
            col.type.geometry_type = "GEOMETRY"

        # Add the SRID of the table in 'gpkg_spatial_ref_sys' if this table exists
        if not bind.execute(
            text("SELECT COUNT(*) FROM gpkg_spatial_ref_sys WHERE srs_id = :srid;").bindparams(
                srid=srid
            )
        ).scalar():
            bind.execute(text("SELECT gpkgInsertEpsgSRID(:srid)").bindparams(srid=srid))
    _setup_dummy_type(table, gis_cols)


def after_create(table, bind, **kw):
    """Handle spatial indexes during the after_create event."""
    dialect = bind.dialect

    for col in table.columns:
        # Add the managed Geometry columns with gpkgAddGeometryColumn()
        if _check_spatial_type(col.type, Geometry, dialect):
            col.type = col._actual_type
            del col._actual_type
            dimension = get_col_dim(col)
            has_z = "Z" in dimension
            has_m = "M" in dimension

            bind.execute(
                text(
                    """INSERT INTO gpkg_contents
                    VALUES (
                        :table_name,
                        'features',
                        :table_name,
                        "",
                        strftime('%Y-%m-%dT%H:%M:%fZ', CURRENT_TIMESTAMP),
                        NULL,
                        NULL,
                        NULL,
                        NULL,
                        :srid
                    );"""
                ).bindparams(
                    table_name=table.name,
                    srid=col.type.srid,
                )
            )
            bind.execute(
                text(
                    """INSERT INTO gpkg_geometry_columns
                    VALUES (:table_name, :column_name, :geometry_type, :srid, :has_z, :has_m);"""
                ).bindparams(
                    table_name=table.name,
                    column_name=col.name,
                    geometry_type=col.type.geometry_type,
                    srid=col.type.srid,
                    has_z=has_z,
                    has_m=has_m,
                )
            )
            stmt = select(*_format_select_args(func.gpkgAddGeometryTriggers(table.name, col.name)))
            stmt = stmt.execution_options(autocommit=True)
            bind.execute(stmt)

    for col in table.columns:
        # Add spatial indexes for the Geometry and Geography columns
        # TODO: Check that the Geography type makes sense here
        if (
            _check_spatial_type(col.type, (Geometry, Geography), dialect)
            and col.type.spatial_index is True
        ):
            create_spatial_index(bind, table, col)

    for idx in table.info.pop("_after_create_indexes"):
        table.indexes.add(idx)
        idx.create(bind=bind)


def before_drop(table, bind, **kw):
    """Handle spatial indexes during the before_drop event."""
    dialect, gis_cols, regular_cols = setup_create_drop(table, bind)

    for col in gis_cols:
        # Disable spatial indexes if present
        # TODO: This is useless but if we remove it then the disable_spatial_index should be
        # tested separately
        disable_spatial_index(bind, table, col)

        # Remove metadata from internal tables
        # (this is equivalent to DiscardGeometryColumn but for GeoPackage)
        bind.execute(
            text(
                """DELETE FROM gpkg_extensions
                WHERE LOWER(table_name) = LOWER(:table_name)
                    AND column_name = :column_name;"""
            ).bindparams(
                table_name=table.name,
                column_name=col.name,
            )
        )
        bind.execute(
            text(
                """DELETE FROM gpkg_geometry_columns
                WHERE LOWER(table_name) = LOWER(:table_name)
                    AND column_name = :column_name;"""
            ).bindparams(
                table_name=table.name,
                column_name=col.name,
            )
        )
        bind.execute(
            text(
                """DELETE FROM gpkg_contents
                WHERE LOWER(table_name) = LOWER(:table_name);"""
            ).bindparams(table_name=table.name)
        )


def after_drop(table, bind, **kw):
    """Handle spatial indexes during the after_drop event."""
    table.columns = table.info.pop("_saved_columns")


def _compiles_gpkg(cls, fn):
    def _compile_gpkg(element, compiler, **kw):
        return "{}({})".format(fn, compiler.process(element.clauses, **kw))

    compiles(getattr(functions, cls), "geopackage")(_compile_gpkg)


def register_gpkg_mapping(mapping):
    """Register compilation mappings for the given functions.

    Args:
        mapping: Should have the following form::

                {
                    "function_name_1": "gpkg_function_name_1",
                    "function_name_2": "gpkg_function_name_2",
                    ...
                }
    """
    for cls, fn in mapping.items():
        _compiles_gpkg(cls, fn)


register_gpkg_mapping(_SQLITE_FUNCTIONS)


def create_spatial_ref_sys_view(bind):
    """Create the `spatial_ref_sys` view from the `gpkg_spatial_ref_sys` table.

    .. Note::

        This is usually only needed to use the `ST_Transform` function on GeoPackage data
        because this function, when used with SpatiaLite, requires the `spatial_ref_sys` table.
    """
    bind.execute(
        text(
            """CREATE VIEW spatial_ref_sys AS
            SELECT
                srs_id AS srid,
                organization AS auth_name,
                organization_coordsys_id AS auth_srid,
                definition AS srtext
            FROM gpkg_spatial_ref_sys;"""
        )
    )



"""Some utils for the GeoAlchemy2 package."""


def authorized_values_in_docstring(**kwargs):
    """Decorator to replace keywords in docstrings by the actual value of a variable.

    .. Note::
        The keyword must be enclose by <> in the docstring, like <MyKeyword>.
    """

    def inner(func):
        if func.__doc__ is not None:
            for k, v in kwargs.items():
                func.__doc__ = func.__doc__.replace(f"<{k}>", str(v))
        return func

    return inner



"""This module defines some dialect-specific functions used for administration tasks."""
from geoalchemy2.admin.dialects import common  # noqa
from geoalchemy2.admin.dialects import geopackage  # noqa
from geoalchemy2.admin.dialects import mysql  # noqa
from geoalchemy2.admin.dialects import postgresql  # noqa
from geoalchemy2.admin.dialects import sqlite  # noqa



"""This module defines specific functions for SQLite dialect."""
import os
from typing import Optional

from sqlalchemy import text
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import func
from sqlalchemy.sql import select

from geoalchemy2 import functions
from geoalchemy2.admin.dialects.common import _check_spatial_type
from geoalchemy2.admin.dialects.common import _format_select_args
from geoalchemy2.admin.dialects.common import _spatial_idx_name
from geoalchemy2.admin.dialects.common import setup_create_drop
from geoalchemy2.types import Geography
from geoalchemy2.types import Geometry
from geoalchemy2.types import _DummyGeometry
from geoalchemy2.utils import authorized_values_in_docstring


def load_spatialite_driver(dbapi_conn, *args):
    """Load SpatiaLite extension in SQLite connection.

    .. Warning::
        The path to the SpatiaLite module should be set in the `SPATIALITE_LIBRARY_PATH`
        environment variable.

    Args:
        dbapi_conn: The DBAPI connection.
    """
    if "SPATIALITE_LIBRARY_PATH" not in os.environ:
        raise RuntimeError("The SPATIALITE_LIBRARY_PATH environment variable is not set.")
    dbapi_conn.enable_load_extension(True)
    dbapi_conn.load_extension(os.environ["SPATIALITE_LIBRARY_PATH"])
    dbapi_conn.enable_load_extension(False)


_JOURNAL_MODE_VALUES = ["DELETE", "TRUNCATE", "PERSIST", "MEMORY", "WAL", "OFF"]


@authorized_values_in_docstring(JOURNAL_MODE_VALUES=_JOURNAL_MODE_VALUES)
def init_spatialite(
    dbapi_conn,
    *args,
    transaction: bool = False,
    init_mode: Optional[str] = None,
    journal_mode: Optional[str] = None,
):
    """Initialize internal SpatiaLite tables.

    Args:
        dbapi_conn: The DBAPI connection.
        init_mode: Can be `None` to load all EPSG SRIDs, `'WGS84'` to load only the ones related
            to WGS84 or `'EMPTY'` to not load any EPSG SRID.

            .. Note::

                It is possible to load other EPSG SRIDs afterwards using `InsertEpsgSrid(srid)`.

        transaction: If set to `True` the whole operation will be handled as a single Transaction
            (faster). The default value is `False` (slower, but safer).
        journal_mode: Change the journal mode to the given value. This can make the table creation
            much faster. The possible values are the following: <JOURNAL_MODE_VALUES>. See
            https://www.sqlite.org/pragma.html#pragma_journal_mode for more details.

            .. Warning::
                Some values, like 'MEMORY' or 'OFF', can lead to corrupted databases if the process
                is interrupted during initialization.

            .. Note::
                The original value is restored after the initialization.

    .. Note::
        When using this function as a listener it is not possible to pass the `transaction`,
        `init_mode` or `journal_mode` arguments directly. To do this you can either create another
        function that calls `init_spatialite` (or
        :func:`geoalchemy2.admin.dialects.sqlite.load_spatialite` if you also want to load the
        SpatiaLite drivers) with an hard-coded `init_mode` or just use a lambda::

            >>> sqlalchemy.event.listen(
            ...     engine,
            ...     "connect",
            ...     lambda x, y: init_spatialite(
            ...         x,
            ...         y,
            ...         transaction=True,
            ...         init_mode="EMPTY",
            ...         journal_mode="OFF",
            ...     )
            ... )
    """
    func_args = []

    # Check the value of the 'transaction' parameter
    if not isinstance(transaction, (bool, int)):
        raise ValueError("The 'transaction' argument must be True or False.")
    else:
        func_args.append(str(transaction))

    # Check the value of the 'init_mode' parameter
    init_mode_values = ["WGS84", "EMPTY"]
    if isinstance(init_mode, str):
        init_mode = init_mode.upper()
    if init_mode is not None:
        if init_mode not in init_mode_values:
            raise ValueError("The 'init_mode' argument must be one of {}.".format(init_mode_values))
        func_args.append(f"'{init_mode}'")

    # Check the value of the 'journal_mode' parameter
    if isinstance(journal_mode, str):
        journal_mode = journal_mode.upper()
    if journal_mode is not None:
        if journal_mode not in _JOURNAL_MODE_VALUES:
            raise ValueError(
                "The 'journal_mode' argument must be one of {}.".format(_JOURNAL_MODE_VALUES)
            )

    if dbapi_conn.execute("SELECT CheckSpatialMetaData();").fetchone()[0] < 1:
        if journal_mode is not None:
            current_journal_mode = dbapi_conn.execute("PRAGMA journal_mode").fetchone()[0]
            dbapi_conn.execute("PRAGMA journal_mode = {}".format(journal_mode))

        dbapi_conn.execute("SELECT InitSpatialMetaData({});".format(", ".join(func_args)))

        if journal_mode is not None:
            dbapi_conn.execute("PRAGMA journal_mode = {}".format(current_journal_mode))


def load_spatialite(*args, **kwargs):
    """Load SpatiaLite extension in SQLite DB and initialize internal tables.

    See :func:`geoalchemy2.admin.dialects.sqlite.load_spatialite_driver` and
    :func:`geoalchemy2.admin.dialects.sqlite.init_spatialite` functions for details about
    arguments.
    """
    load_spatialite_driver(*args)
    init_spatialite(*args, **kwargs)


def _get_spatialite_attrs(bind, table_name, col_name):
    attrs = bind.execute(
        text(
            """SELECT * FROM "geometry_columns"
            WHERE LOWER(f_table_name) = LOWER(:table_name)
                AND LOWER(f_geometry_column) = LOWER(:column_name)
        """
        ).bindparams(table_name=table_name, column_name=col_name)
    ).fetchone()
    if attrs is None:
        # If the column is not registered as a spatial column we ignore it
        return None
    return attrs[2:]


def get_spatialite_version(bind):
    """Get the version of the currently loaded Spatialite extension."""
    return bind.execute(text("SELECT spatialite_version();")).fetchone()[0]


def _setup_dummy_type(table, gis_cols):
    """Setup dummy type for new Geometry columns so they can be updated later."""
    for col in gis_cols:
        # Add dummy columns with GEOMETRY type
        col._actual_type = col.type
        col.type = _DummyGeometry()
    table.columns = table.info["_saved_columns"]


def get_col_dim(col):
    """Get dimension of the column type."""
    if col.type.dimension == 4:
        dimension = "XYZM"
    elif col.type.dimension == 2:
        dimension = "XY"
    else:
        if col.type.geometry_type.endswith("M"):
            dimension = "XYM"
        else:
            dimension = "XYZ"
    return dimension


def create_spatial_index(bind, table, col):
    """Create spatial index on the given column."""
    stmt = select(*_format_select_args(func.CreateSpatialIndex(table.name, col.name)))
    stmt = stmt.execution_options(autocommit=True)
    bind.execute(stmt)


def disable_spatial_index(bind, table, col):
    """Disable spatial indexes if present."""
    stmt = select(*_format_select_args(func.CheckSpatialIndex(table.name, col.name)))
    if bind.execute(stmt).fetchone()[0] is not None:
        stmt = select(*_format_select_args(func.DisableSpatialIndex(table.name, col.name)))
        stmt = stmt.execution_options(autocommit=True)
        bind.execute(stmt)
        bind.execute(
            text(
                "DROP TABLE IF EXISTS {};".format(
                    _spatial_idx_name(
                        table.name,
                        col.name,
                    )
                )
            )
        )


def reflect_geometry_column(inspector, table, column_info):
    """Reflect a column of type Geometry with SQLite dialect."""
    # Get geometry type, SRID and spatial index from the SpatiaLite metadata
    if not isinstance(column_info.get("type"), Geometry):
        return
    col_attributes = _get_spatialite_attrs(inspector.bind, table.name, column_info["name"])
    if col_attributes is not None:
        geometry_type, coord_dimension, srid, spatial_index = col_attributes

        if isinstance(geometry_type, int):
            geometry_type_str = str(geometry_type)
            if geometry_type >= 1000:
                first_digit = geometry_type_str[0]
                has_z = first_digit in ["1", "3"]
                has_m = first_digit in ["2", "3"]
            else:
                has_z = has_m = False
            geometry_type = {
                "0": "GEOMETRY",
                "1": "POINT",
                "2": "LINESTRING",
                "3": "POLYGON",
                "4": "MULTIPOINT",
                "5": "MULTILINESTRING",
                "6": "MULTIPOLYGON",
                "7": "GEOMETRYCOLLECTION",
            }[geometry_type_str[-1]]
            if has_z:
                geometry_type += "Z"
            if has_m:
                geometry_type += "M"
        else:
            if "Z" in coord_dimension and "Z" not in geometry_type[-2:]:
                geometry_type += "Z"
            if "M" in coord_dimension and "M" not in geometry_type[-2:]:
                geometry_type += "M"
            coord_dimension = {
                "XY": 2,
                "XYZ": 3,
                "XYM": 3,
                "XYZM": 4,
            }.get(coord_dimension, coord_dimension)

        # Set attributes
        column_info["type"].geometry_type = geometry_type
        column_info["type"].dimension = coord_dimension
        column_info["type"].srid = srid
        column_info["type"].spatial_index = bool(spatial_index)

        # Spatial indexes are not automatically reflected with SQLite dialect
        column_info["type"]._spatial_index_reflected = False


def before_create(table, bind, **kw):
    """Handle spatial indexes during the before_create event."""
    dialect, gis_cols, regular_cols = setup_create_drop(table, bind)

    # Remove the spatial indexes from the table metadata because they should not be
    # created during the table.create() step since the associated columns do not exist
    # at this time.
    table.info["_after_create_indexes"] = []
    current_indexes = set(table.indexes)
    for idx in current_indexes:
        for col in table.info["_saved_columns"]:
            if (_check_spatial_type(col.type, Geometry, dialect)) and col in idx.columns.values():
                table.indexes.remove(idx)
                if idx.name != _spatial_idx_name(table.name, col.name) or not getattr(
                    col.type, "spatial_index", False
                ):
                    table.info["_after_create_indexes"].append(idx)

    _setup_dummy_type(table, gis_cols)


def after_create(table, bind, **kw):
    """Handle spatial indexes during the after_create event."""
    dialect = bind.dialect

    table.columns = table.info.pop("_saved_columns")
    for col in table.columns:
        # Add the managed Geometry columns with RecoverGeometryColumn()
        if _check_spatial_type(col.type, Geometry, dialect):
            col.type = col._actual_type
            del col._actual_type
            dimension = get_col_dim(col)
            args = [table.name, col.name, col.type.srid, col.type.geometry_type, dimension]

            stmt = select(*_format_select_args(func.RecoverGeometryColumn(*args)))
            stmt = stmt.execution_options(autocommit=True)
            bind.execute(stmt)

    for col in table.columns:
        # Add spatial indexes for the Geometry and Geography columns
        # TODO: Check that the Geography type makes sense here
        if (
            _check_spatial_type(col.type, (Geometry, Geography), dialect)
            and col.type.spatial_index is True
        ):
            create_spatial_index(bind, table, col)

    for idx in table.info.pop("_after_create_indexes"):
        table.indexes.add(idx)
        idx.create(bind=bind)


def before_drop(table, bind, **kw):
    """Handle spatial indexes during the before_drop event."""
    dialect, gis_cols, regular_cols = setup_create_drop(table, bind)

    for col in gis_cols:
        # Disable spatial indexes if present
        disable_spatial_index(bind, table, col)

        args = [table.name, col.name]

        stmt = select(*_format_select_args(func.DiscardGeometryColumn(*args)))
        stmt = stmt.execution_options(autocommit=True)
        bind.execute(stmt)


def after_drop(table, bind, **kw):
    """Handle spatial indexes during the after_drop event."""
    table.columns = table.info.pop("_saved_columns")


# Define compiled versions for functions in SpatiaLite whose names don't have
# the ST_ prefix.
_SQLITE_FUNCTIONS = {
    "ST_GeomFromEWKT": "GeomFromEWKT",
    "ST_GeomFromEWKB": "GeomFromEWKB",
    "ST_AsBinary": "AsBinary",
    "ST_AsEWKB": "AsEWKB",
    "ST_AsGeoJSON": "AsGeoJSON",
}


def _compiles_sqlite(cls, fn):
    def _compile_sqlite(element, compiler, **kw):
        return "{}({})".format(fn, compiler.process(element.clauses, **kw))

    compiles(getattr(functions, cls), "sqlite")(_compile_sqlite)


def register_sqlite_mapping(mapping):
    """Register compilation mappings for the given functions.

    Args:
        mapping: Should have the following form::

                {
                    "function_name_1": "sqlite_function_name_1",
                    "function_name_2": "sqlite_function_name_2",
                    ...
                }
    """
    for cls, fn in mapping.items():
        _compiles_sqlite(cls, fn)


register_sqlite_mapping(_SQLITE_FUNCTIONS)



from pathlib import Path

from geoalchemy2._functions_helpers import _generate_stubs

"""
this script is outside the geoalchemy2 package because the 'geoalchemy2.types'
package interferes with the 'types' module in the standard library
"""

script_dir = Path(__file__).resolve().parent


if __name__ == "__main__":
    (script_dir / "geoalchemy2/functions.pyi").write_text(_generate_stubs())



"""
New ORM Declarative Mapping Style
=================================

``SQLAlchemy>=2`` introduced a new way to construct mappings using the
``sqlalchemy.orm.DeclarativeBase`` base class.
This example shows how to use GeoAlchemy2 types in this context.
"""
import pytest
from pkg_resources import parse_version
from sqlalchemy import __version__ as SA_VERSION

try:
    from sqlalchemy.orm import DeclarativeBase
    from sqlalchemy.orm import Mapped
    from sqlalchemy.orm import mapped_column
except ImportError:
    pass

from geoalchemy2 import Geometry
from geoalchemy2 import WKBElement
from geoalchemy2 import shape


def check_wkb(wkb, x, y) -> None:
    pt = shape.to_shape(wkb)
    assert round(pt.x, 5) == x
    assert round(pt.y, 5) == y


@pytest.mark.skipif(
    parse_version(SA_VERSION) < parse_version("2"),
    reason="New ORM mapping is only available for sqlalchemy>=2",
)
def test_ORM_mapping(session, conn, schema) -> None:
    class Base(DeclarativeBase):
        pass

    class Lake(Base):
        __tablename__ = "lake"
        __table_args__ = {"schema": schema}
        id: Mapped[int] = mapped_column(primary_key=True)
        mapped_geom: Mapped[WKBElement] = mapped_column(Geometry(geometry_type="POINT", srid=4326))

    Lake.__table__.drop(conn, checkfirst=True)  # type: ignore[attr-defined]
    Lake.__table__.create(bind=conn)  # type: ignore[attr-defined]

    # Create new point instance
    p = Lake()
    p.mapped_geom = "SRID=4326;POINT(5 45)"  # type: ignore[assignment]

    # Insert point
    session.add(p)
    session.flush()
    session.expire(p)

    # Query the point and check the result
    pt = session.query(Lake).one()
    assert pt.id == 1
    assert pt.mapped_geom.srid == 4326
    check_wkb(pt.mapped_geom, 5, 45)
